""" Client for interacting with Framework/BundleDelivery service
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import getpass
import tarfile
from six import BytesIO

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.DISET.TransferClient import TransferClient
from DIRAC.Core.Security import Locations, Utilities
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.ConfigurationSystem.Client.Helpers.CSGlobals import skipCACheck


__RCSID__ = "$Id$"


@createClient("Framework/BundleDelivery")
class BundleDeliveryClient(Client):
    def __init__(self, transferClient=False, **kwargs):
        super(BundleDeliveryClient, self).__init__(**kwargs)
        self.setServer("Framework/BundleDelivery")
        self.transferClient = transferClient
        self.log = gLogger.getSubLogger("BundleDelivery")

    def __getTransferClient(self):
        """Get transfer client

        :return: TransferClient()
        """
        if self.transferClient:
            return self.transferClient
        return TransferClient("Framework/BundleDelivery", skipCACheck=skipCACheck())

    def __getHash(self, bundleID, dirToSyncTo):
        """Get hash for bundle in directory

        :param str bundleID: bundle ID
        :param str dirToSyncTo: path to sync directory

        :return: str
        """
        try:
            with open(os.path.join(dirToSyncTo, ".dab.%s" % bundleID), "r") as fd:
                bdHash = fd.read().strip()
                return bdHash
        except Exception:
            return ""

    def __setHash(self, bundleID, dirToSyncTo, bdHash):
        """Set hash for bundle in directory

        :param str bundleID: bundle ID
        :param str dirToSyncTo: path to sync directory
        :param str bdHash: new hash
        """
        try:
            fileName = os.path.join(dirToSyncTo, ".dab.%s" % bundleID)
            with open(fileName, "wt") as fd:
                fd.write(bdHash)
        except Exception as e:
            self.log.error("Could not save hash after synchronization", "%s: %s" % (fileName, str(e)))

    def syncDir(self, bundleID, dirToSyncTo):
        """Synchronize directory

        :param str bundleID: bundle ID
        :param str dirToSyncTo: path to sync directory

        :return: S_OK(bool)/S_ERROR()
        """
        dirCreated = False
        if os.path.isdir(dirToSyncTo):
            for p in [os.W_OK, os.R_OK]:
                if not os.access(dirToSyncTo, p):
                    self.log.error("%s does not have the permissions to update %s" % (getpass.getuser(), dirToSyncTo))
                    return S_ERROR("%s does not have the permissions to update %s" % (getpass.getuser(), dirToSyncTo))
        else:
            self.log.info("Creating dir %s" % dirToSyncTo)
            mkDir(dirToSyncTo)
            dirCreated = True
        currentHash = self.__getHash(bundleID, dirToSyncTo)
        self.log.info("Current hash for bundle %s in dir %s is '%s'" % (bundleID, dirToSyncTo, currentHash))
        buff = BytesIO()
        transferClient = self.__getTransferClient()
        result = transferClient.receiveFile(buff, [bundleID, currentHash])
        if not result["OK"]:
            self.log.error("Could not sync dir", result["Message"])
            if dirCreated:
                self.log.info("Removing dir %s" % dirToSyncTo)
                os.unlink(dirToSyncTo)
            buff.close()
            return result
        newHash = result["Value"]
        if newHash == currentHash:
            self.log.info("Dir %s was already in sync" % dirToSyncTo)
            return S_OK(False)
        buff.seek(0)
        self.log.info("Synchronizing dir with remote bundle")
        with tarfile.open(name="dummy", mode="r:gz", fileobj=buff) as tF:
            for tarinfo in tF:
                try:
                    tF.extract(tarinfo, dirToSyncTo)
                except OSError as e:
                    self.log.error("Could not sync dir:", str(e))
                    if dirCreated:
                        self.log.info("Removing dir %s" % dirToSyncTo)
                        os.unlink(dirToSyncTo)
                    buff.close()
                    return S_ERROR("Certificates directory update failed: %s" % str(e))

        buff.close()
        self.__setHash(bundleID, dirToSyncTo, newHash)
        self.log.info("Dir has been synchronized")
        return S_OK(True)

    def syncCAs(self):
        """Synchronize CAs

        :return: S_OK(bool)/S_ERROR()
        """
        X509_CERT_DIR = False
        if "X509_CERT_DIR" in os.environ:
            X509_CERT_DIR = os.environ["X509_CERT_DIR"]
            del os.environ["X509_CERT_DIR"]
        casLocation = Locations.getCAsLocation()
        if not casLocation:
            casLocation = Locations.getCAsDefaultLocation()
        result = self.syncDir("CAs", casLocation)
        if X509_CERT_DIR:
            os.environ["X509_CERT_DIR"] = X509_CERT_DIR
        return result

    def syncCRLs(self):
        """Synchronize CRLs

        :return: S_OK(bool)/S_ERROR()
        """
        X509_CERT_DIR = False
        if "X509_CERT_DIR" in os.environ:
            X509_CERT_DIR = os.environ["X509_CERT_DIR"]
            del os.environ["X509_CERT_DIR"]
        result = self.syncDir("CRLs", Locations.getCAsLocation())
        if X509_CERT_DIR:
            os.environ["X509_CERT_DIR"] = X509_CERT_DIR
        return result

    def getCAs(self):
        """This method can be used to create the CAs. If the file can not be created,
        it will be downloaded from the server.

        :return: S_OK(str)/S_ERROR()
        """
        retVal = Utilities.generateCAFile()
        if not retVal["OK"]:
            self.log.warn("Could not generate/find CA file", retVal["Message"])
            # if we can not found the file, we return the directory, where the file should be
            transferClient = self.__getTransferClient()
            casFile = os.path.join(os.path.dirname(retVal["Message"]), "cas.pem")
            with open(casFile, "w") as fd:
                result = transferClient.receiveFile(fd, "CAs")
                if not result["OK"]:
                    return result
                return S_OK(casFile)
        else:
            return retVal

    def getCLRs(self):
        """This method can be used to create the CRLs. If the file can not be created,
        it will be downloaded from the server.

        :return: S_OK(str)/S_ERROR()
        """
        retVal = Utilities.generateRevokedCertsFile()
        if not retVal["OK"]:
            # if we can not found the file, we return the directory, where the file should be
            transferClient = self.__getTransferClient()
            casFile = os.path.join(os.path.dirname(retVal["Message"]), "crls.pem")
            with open(casFile, "w") as fd:
                result = transferClient.receiveFile(fd, "CRLs")
                if not result["OK"]:
                    return result
                return S_OK(casFile)
        else:
            return retVal



########################################################################
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2012/12/11 18:04:25
########################################################################

""" :mod: SubprocessTests
    =======================

    .. module: SubprocessTests
    :synopsis: unittest for Subprocess module
    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    unittest for Subprocess module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from os.path import dirname, join

# imports
import time
import pytest

from subprocess import Popen

# SUT
from DIRAC.Core.Utilities.Subprocess import systemCall, shellCall, pythonCall, getChildrenPIDs, Subprocess

# Mark this entire module as slow
pytestmark = pytest.mark.slow

cmd = ["sleep", "2"]


def pyfunc(_name):
    time.sleep(2)


@pytest.mark.parametrize("timeout, expected", [(False, True), (3, True), (1, False)])
def test_calls(timeout, expected):
    ret = systemCall(timeout, cmdSeq=cmd)
    assert ret["OK"] == expected

    ret = shellCall(timeout, cmdSeq=" ".join(cmd))
    assert ret["OK"] == expected

    ret = pythonCall(timeout, pyfunc, "something")
    assert ret["OK"] == expected


def test_getChildrenPIDs():
    import os

    os.system("echo $PWD")
    mainProcess = Popen(["python", join(dirname(__file__), "ProcessesCreator.py")])
    time.sleep(1)
    res = getChildrenPIDs(mainProcess.pid)
    assert len(res) == 3
    for p in res:
        assert isinstance(p, int)

    mainProcess.wait()


def test_decodingCommandOutput():
    retVal = systemCall(10, ["echo", "-e", "-n", r"\xdf"])
    assert retVal["OK"]
    assert retVal["Value"] == (0, u"\ufffd", "")

    retVal = systemCall(10, ["echo", "-e", r"\xdf"])
    assert retVal["OK"]
    assert retVal["Value"] == (0, u"\ufffd\n", "")

    sp = Subprocess()
    retVal = sp.systemCall(r"""python -c 'import os; os.fdopen(2, "wb").write(b"\xdf")'""", shell=True)
    assert retVal["OK"]
    assert retVal["Value"] == (0, "", u"\ufffd")



""" Class that contains client access to the transformation DB handler. """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.TransformationSystem.Client import TransformationFilesStatus


@createClient("Transformation/TransformationManager")
class TransformationClient(Client):
    """Exposes the functionality available in the DIRAC/TransformationManagerHandler

    This inherits the DIRAC base Client for direct execution of server functionality.
    The following methods are available (although not visible here).

    Transformation (table) manipulation

        deleteTransformation(transName)
        getTransformationParameters(transName,paramNames)
        getTransformationWithStatus(status)
        setTransformationParameter(transName,paramName,paramValue)
        deleteTransformationParameter(transName,paramName)

    TransformationFiles table manipulation

        addFilesToTransformation(transName,lfns)
        addTaskForTransformation(transName,lfns=[],se='Unknown')
        getTransformationStats(transName)

    TransformationTasks table manipulation

        setTaskStatus(transName, taskID, status)
        setTaskStatusAndWmsID(transName, taskID, status, taskWmsID)
        getTransformationTaskStats(transName)
        deleteTasks(transName, taskMin, taskMax)
        extendTransformation( transName, nTasks)
        getTasksToSubmit(transName,numTasks,site='')

    TransformationLogging table manipulation

        getTransformationLogging(transName)

    File/directory manipulation methods (the remainder of the interface can be found below)

        getFileSummary(lfns)
        exists(lfns)

    Web monitoring tools

        getDistinctAttributeValues(attribute, selectDict)
        getTransformationStatusCounters()
        getTransformationSummary()
        getTransformationSummaryWeb(selectDict, sortList, startItem, maxItems)
    """

    def __init__(self, **kwargs):
        """Simple constructor"""

        super(TransformationClient, self).__init__(**kwargs)
        opsH = Operations()
        self.maxResetCounter = opsH.getValue("Transformations/FilesMaxResetCounter", 10)

        self.setServer("Transformation/TransformationManager")

    def setServer(self, url):
        self.serverURL = url

    def getCounters(self, table, attrList, condDict, older=None, newer=None, timeStamp=None):
        rpcClient = self._getRPC()
        return rpcClient.getCounters(table, attrList, condDict, older, newer, timeStamp)

    def addTransformation(
        self,
        transName,
        description,
        longDescription,
        transType,
        plugin,
        agentType,
        fileMask,
        transformationGroup="General",
        groupSize=1,
        inheritedFrom=0,
        body="",
        maxTasks=0,
        eventsPerTask=0,
        addFiles=True,
        inputMetaQuery=None,
        outputMetaQuery=None,
        timeout=1800,
    ):
        """add a new transformation"""
        rpcClient = self._getRPC(timeout=timeout)
        return rpcClient.addTransformation(
            transName,
            description,
            longDescription,
            transType,
            plugin,
            agentType,
            fileMask,
            transformationGroup,
            groupSize,
            inheritedFrom,
            body,
            maxTasks,
            eventsPerTask,
            addFiles,
            inputMetaQuery,
            outputMetaQuery,
        )

    def getTransformations(
        self,
        condDict=None,
        older=None,
        newer=None,
        timeStamp=None,
        orderAttribute=None,
        limit=100,
        extraParams=False,
        columns=None,
    ):
        """gets all the transformations in the system, incrementally. "limit" here is just used to determine the offset."""
        rpcClient = self._getRPC()

        transformations = []
        if condDict is None:
            condDict = {}
        if timeStamp is None:
            timeStamp = "CreationDate"
        # getting transformations - incrementally
        offsetToApply = 0
        while True:
            res = rpcClient.getTransformations(
                condDict, older, newer, timeStamp, orderAttribute, limit, extraParams, offsetToApply, columns
            )
            if not res["OK"]:
                return res
            else:
                gLogger.verbose("Result for limit %d, offset %d: %d" % (limit, offsetToApply, len(res["Value"])))
                if res["Value"]:
                    transformations = transformations + res["Value"]
                    offsetToApply += limit
                if len(res["Value"]) < limit:
                    break
        return S_OK(transformations)

    def getTransformation(self, transName, extraParams=False):
        rpcClient = self._getRPC()
        return rpcClient.getTransformation(transName, extraParams)

    def getTransformationFiles(
        self,
        condDict=None,
        older=None,
        newer=None,
        timeStamp=None,
        orderAttribute=None,
        limit=None,
        timeout=1800,
        offset=0,
        maxfiles=None,
    ):
        """gets all the transformation files for a transformation, incrementally.
        "limit" here is just used to determine the offset.
        """
        rpcClient = self._getRPC(timeout=timeout)
        transformationFiles = []
        if condDict is None:
            condDict = {}
        if timeStamp is None:
            timeStamp = "LastUpdate"
        # getting transformationFiles - incrementally
        if "LFN" in condDict:
            if isinstance(condDict["LFN"], six.string_types):
                lfnList = [condDict["LFN"]]
            else:
                lfnList = sorted(condDict["LFN"])
            # If a list of LFNs is given, use chunks of 1000 only
            limit = limit if limit else 1000
        else:
            # By default get by chunks of 10000 files
            lfnList = []
            limit = limit if limit else 10000
        transID = condDict.get("TransformationID", "Unknown")
        offsetToApply = offset
        retries = 5
        while True:
            if lfnList:
                # If list is exhausted, exit
                if offsetToApply >= len(lfnList):
                    break
                # Apply the offset to the list of LFNs
                condDict["LFN"] = lfnList[offsetToApply : offsetToApply + limit]
                # No limit and no offset as the list is limited already
                res = rpcClient.getTransformationFiles(condDict, older, newer, timeStamp, orderAttribute, None, None)
            else:
                res = rpcClient.getTransformationFiles(
                    condDict, older, newer, timeStamp, orderAttribute, limit, offsetToApply
                )
            if not res["OK"]:
                gLogger.error(
                    "Error getting files for transformation %s (offset %d), %s"
                    % (str(transID), offsetToApply, ("retry %d times" % retries) if retries else "give up"),
                    res["Message"],
                )
                retries -= 1
                if retries:
                    continue
                return res
            else:
                condDictStr = str(condDict)
                log = gLogger.debug if len(condDictStr) > 100 else gLogger.verbose
                if not log(
                    "For conditions %s: result for limit %d, offset %d: %d files"
                    % (condDictStr, limit, offsetToApply, len(res["Value"]))
                ):
                    gLogger.verbose(
                        "For condition keys %s (trans %s): result for limit %d, offset %d: %d files"
                        % (
                            str(sorted(condDict)),
                            condDict.get("TransformationID", "None"),
                            limit,
                            offsetToApply,
                            len(res["Value"]),
                        )
                    )
                if res["Value"]:
                    transformationFiles += res["Value"]
                    # Limit the number of files returned
                    if maxfiles and len(transformationFiles) >= maxfiles:
                        transformationFiles = transformationFiles[:maxfiles]
                        break
                # Less data than requested, exit only if LFNs were not given
                if not lfnList and len(res["Value"]) < limit:
                    break
                offsetToApply += limit
                # Reset number of retries for next chunk
                retries = 5

        return S_OK(transformationFiles)

    def getTransformationTasks(
        self, condDict=None, older=None, newer=None, timeStamp=None, orderAttribute=None, limit=10000, inputVector=False
    ):
        """gets all the transformation tasks for a transformation, incrementally.
        "limit" here is just used to determine the offset.
        """
        rpcClient = self._getRPC()
        transformationTasks = []
        if condDict is None:
            condDict = {}
        if timeStamp is None:
            timeStamp = "CreationTime"
        # getting transformationFiles - incrementally
        offsetToApply = 0
        while True:
            res = rpcClient.getTransformationTasks(
                condDict, older, newer, timeStamp, orderAttribute, limit, inputVector, offsetToApply
            )
            if not res["OK"]:
                return res
            else:
                gLogger.verbose("Result for limit %d, offset %d: %d" % (limit, offsetToApply, len(res["Value"])))
                if res["Value"]:
                    transformationTasks = transformationTasks + res["Value"]
                    offsetToApply += limit
                if len(res["Value"]) < limit:
                    break
        return S_OK(transformationTasks)

    def cleanTransformation(self, transID):
        """Clean the transformation, and set the status parameter (doing it here, for easier extensibility)"""
        # Cleaning
        rpcClient = self._getRPC()
        res = rpcClient.cleanTransformation(transID)
        if not res["OK"]:
            return res
        # Setting the status
        return self.setTransformationParameter(transID, "Status", "TransformationCleaned")

    # Add methods to handle transformation status

    def startTransformation(self, transID):
        """Start the transformation"""
        res = self.setTransformationParameter(transID, "Status", "Active")
        if not res["OK"]:
            gLogger.error("Failed to start transformation %s: %s" % (transID, res["Message"]))
            return res
        else:
            res = self.setTransformationParameter(transID, "AgentType", "Automatic")
            if not res["OK"]:
                gLogger.error("Failed to set AgentType to transformation %s: %s" % (transID, res["Message"]))

        return res

    def stopTransformation(self, transID):
        """Stop the transformation"""
        res = self.setTransformationParameter(transID, "Status", "Stopped")
        if not res["OK"]:
            gLogger.error("Failed to stop transformation %s: %s" % (transID, res["Message"]))
            return res
        else:
            res = self.setTransformationParameter(transID, "AgentType", "Manual")
            if not res["OK"]:
                gLogger.error("Failed to set AgentType to transformation %s: %s" % (transID, res["Message"]))

        return res

    def moveFilesToDerivedTransformation(self, transDict, resetUnused=True):
        """move files input to a transformation, to the derived one"""
        prod = transDict["TransformationID"]
        parentProd = int(transDict.get("InheritedFrom", 0))
        movedFiles = {}
        log = gLogger.getLocalSubLogger("[None] [%d] .moveFilesToDerivedTransformation:" % prod)
        if not parentProd:
            log.warn("Transformation was not derived...")
            return S_OK((parentProd, movedFiles))
        # get the lfns in status Unused/MaxReset of the parent production
        res = self.getTransformationFiles(
            condDict={
                "TransformationID": parentProd,
                "Status": [TransformationFilesStatus.UNUSED, TransformationFilesStatus.MAX_RESET],
            }
        )
        if not res["OK"]:
            log.error(" Error getting Unused files from transformation", "%d: %s" % (parentProd, res["Message"]))
            return res
        parentFiles = res["Value"]
        lfns = [lfnDict["LFN"] for lfnDict in parentFiles]
        if not lfns:
            log.info(" No files found to be moved from transformation", "%d" % parentProd)
            return S_OK((parentProd, movedFiles))
        # get the lfns of the derived production that were Unused/MaxReset in the parent one
        res = self.getTransformationFiles(condDict={"TransformationID": prod, "LFN": lfns})
        if not res["OK"]:
            log.error(" Error getting files from derived transformation:", res["Message"])
            return res
        derivedFiles = res["Value"]
        derivedStatusDict = dict((derivedDict["LFN"], derivedDict["Status"]) for derivedDict in derivedFiles)
        newStatusFiles = {}
        parentStatusFiles = {}
        badStatusFiles = {}
        for parentDict in parentFiles:
            lfn = parentDict["LFN"]
            derivedStatus = derivedStatusDict.get(lfn)
            if derivedStatus:
                parentStatus = parentDict["Status"]
                # By default move to the parent status (which is Unused or MaxReset)
                status = parentStatus
                moveStatus = parentStatus
                # For MaxReset, set Unused if requested
                if parentStatus == TransformationFilesStatus.MAX_RESET:
                    if resetUnused:
                        status = TransformationFilesStatus.UNUSED
                        moveStatus = "Unused from MaxReset"
                    else:
                        status = "MaxReset-inherited"
                if derivedStatus.endswith("-inherited"):
                    # This is the general case
                    newStatusFiles.setdefault((status, parentStatus), []).append(lfn)
                    movedFiles[moveStatus] = movedFiles.setdefault(moveStatus, 0) + 1
                else:
                    badStatusFiles[derivedStatus] = badStatusFiles.setdefault(derivedStatus, 0) + 1
                if parentStatus == TransformationFilesStatus.UNUSED:
                    # If the file was Unused, set it NotProcessed in parent
                    parentStatusFiles.setdefault("NotProcessed", []).append(lfn)
                else:
                    parentStatusFiles.setdefault("Moved", []).append(lfn)

        for status, count in badStatusFiles.items():  # can be an iterator
            log.warn(
                "Files found in an unexpected status in derived transformation",
                ": %d files in status %s" % (count, status),
            )
        # Set the status in the parent transformation first
        for status, lfnList in parentStatusFiles.items():  # can be an iterator
            for lfnChunk in breakListIntoChunks(lfnList, 5000):
                res = self.setFileStatusForTransformation(parentProd, status, lfnChunk)
                if not res["OK"]:
                    log.error(
                        " Error setting status in transformation",
                        "%d: status %s for %d files - %s" % (parentProd, status, len(lfnList), res["Message"]),
                    )

        # Set the status in the new transformation
        for (status, oldStatus), lfnList in newStatusFiles.items():  # can be an iterator
            for lfnChunk in breakListIntoChunks(lfnList, 5000):
                res = self.setFileStatusForTransformation(prod, status, lfnChunk)
                if not res["OK"]:
                    log.debug(
                        " Error setting status in transformation",
                        "%d: status %s for %d files; resetting them %s. %s"
                        % (parentProd, status, len(lfnChunk), oldStatus, res["Message"]),
                    )
                    res = self.setFileStatusForTransformation(parentProd, oldStatus, lfnChunk)
                    if not res["OK"]:
                        log.error(
                            " Error setting status in transformation",
                            " %d: status %s for %d files: %s" % (parentProd, oldStatus, len(lfnChunk), res["Message"]),
                        )
                else:
                    log.info(
                        "Successfully moved files", ": %d files from %s to %s" % (len(lfnChunk), oldStatus, status)
                    )

        # If files were Assigned or Unused at the time of derivation, try and update them as jobs may have run since then
        res = self.getTransformationFiles(
            condDict={
                "TransformationID": prod,
                "Status": [TransformationFilesStatus.ASSIGNED_INHERITED, TransformationFilesStatus.UNUSED_INHERITED],
            }
        )
        if res["OK"]:
            assignedFiles = res["Value"]
            if assignedFiles:
                lfns = [lfnDict["LFN"] for lfnDict in assignedFiles]
                res = self.getTransformationFiles(condDict={"TransformationID": parentProd, "LFN": lfns})
                if res["OK"]:
                    parentFiles = res["Value"]
                    processedLfns = [
                        lfnDict["LFN"]
                        for lfnDict in parentFiles
                        if lfnDict["Status"] == TransformationFilesStatus.PROCESSED
                    ]
                    if processedLfns:
                        res = self.setFileStatusForTransformation(
                            prod, TransformationFilesStatus.PROCESSED_INHERITED, processedLfns
                        )
                        if res["OK"]:
                            log.info(
                                "Successfully set files status",
                                ": %d files to status %s"
                                % (len(processedLfns), TransformationFilesStatus.PROCESSED_INHERITED),
                            )
        if not res["OK"]:
            log.error("Error setting status for Assigned derived files", res["Message"])

        return S_OK((parentProd, movedFiles))

    def setFileStatusForTransformation(self, transName, newLFNsStatus=None, lfns=None, force=False):
        """Sets the file status for LFNs of a transformation

        For backward compatibility purposes, the status and LFNs can be passed in 2 ways:

          - newLFNsStatus is a dictionary with the form:
            {'/this/is/an/lfn1.txt': 'StatusA', '/this/is/an/lfn2.txt': 'StatusB',  ... }
            and at this point lfns is not considered
          - newLFNStatus is a string, that applies to all the LFNs in lfns

        """
        # create dictionary in case newLFNsStatus is a string
        if isinstance(newLFNsStatus, six.string_types):
            if not lfns:
                return S_OK({})
            if isinstance(lfns, six.string_types):
                lfns = [lfns]
            newLFNsStatus = dict.fromkeys(lfns, newLFNsStatus)
        if not newLFNsStatus:
            return S_OK({})

        rpcClient = self._getRPC()
        # gets current status, errorCount and fileID
        tsFiles = self.getTransformationFiles({"TransformationID": transName, "LFN": list(newLFNsStatus)})
        if not tsFiles["OK"]:
            return tsFiles
        tsFiles = tsFiles["Value"]
        newStatuses = {}
        if tsFiles:
            # for convenience, makes a small dictionary out of the tsFiles, with the lfn as key
            tsFilesAsDict = dict(
                (tsFile["LFN"], [tsFile["Status"], tsFile["ErrorCount"], tsFile["FileID"]]) for tsFile in tsFiles
            )

            # applying the state machine to the proposed status
            newStatuses = self._applyTransformationFilesStateMachine(tsFilesAsDict, newLFNsStatus, force)

            if newStatuses:  # if there's something to update
                # Key to the service is fileIDs
                # The value is a tuple with the new status and a flag that says if ErrorCount should be incremented
                newStatusForFileIDs = dict(
                    (
                        tsFilesAsDict[lfn][2],
                        [newStatuses[lfn], self._wasFileInError(newStatuses[lfn], tsFilesAsDict[lfn][0])],
                    )
                    for lfn in newStatuses
                )
                res = rpcClient.setFileStatusForTransformation(transName, newStatusForFileIDs)
                if not res["OK"]:
                    return res

        return S_OK(newStatuses)

    def _wasFileInError(self, newStatus, currentStatus):
        """Tells whether the file was Assigned and failed, i.e. was not Processed"""
        return currentStatus == TransformationFilesStatus.ASSIGNED and newStatus != TransformationFilesStatus.PROCESSED

    def _applyTransformationFilesStateMachine(self, tsFilesAsDict, dictOfProposedLFNsStatus, force):
        """For easier extension, here we apply the state machine of the production files.
        VOs might want to replace the standard here with something they prefer.

        tsFiles is a dictionary with the lfn as key and as value a list of [Status, ErrorCount, FileID]
        dictOfNewLFNsStatus is a dictionary with the proposed status
        force is a boolean

        It returns a dictionary with the status updates
        """
        newStatuses = {}

        for lfn, newStatus in dictOfProposedLFNsStatus.items():  # can be an iterator
            if lfn in tsFilesAsDict:
                currentStatus = tsFilesAsDict[lfn][0]
                # Apply optional corrections
                if (
                    currentStatus == TransformationFilesStatus.PROCESSED
                    and newStatus != TransformationFilesStatus.PROCESSED
                ):
                    # Processed files should be in a final status unless forced
                    if not force:
                        newStatus = TransformationFilesStatus.PROCESSED
                elif currentStatus == TransformationFilesStatus.MAX_RESET:
                    # MaxReset files can go to any status except Unused (unless forced)
                    if newStatus == TransformationFilesStatus.UNUSED and not force:
                        newStatus = TransformationFilesStatus.MAX_RESET
                elif newStatus == TransformationFilesStatus.UNUSED:
                    errorCount = tsFilesAsDict[lfn][1]
                    # every 10 retries (by default) the file cannot be reset Unused any longer
                    if errorCount and ((errorCount % self.maxResetCounter) == 0) and not force:
                        newStatus = TransformationFilesStatus.MAX_RESET

                # Only worth changing status if it is different
                if newStatus != currentStatus:
                    newStatuses[lfn] = newStatus

        return newStatuses

    def setTransformationParameter(self, transID, paramName, paramValue, force=False, currentStatus=None):
        """Sets a transformation parameter. There's a special case when coming to setting the status of a transformation.

        :param currentStatus: if set, make sure the status did not change in the DB before setting it
        """
        rpcClient = self._getRPC()

        if paramName.lower() == "status":
            # get transformation Type
            transformation = self.getTransformation(transID)
            if not transformation["OK"]:
                return transformation
            transformationType = transformation["Value"]["Type"]

            # get status as of today
            originalStatus = self.getTransformationParameters(transID, "Status")
            if not originalStatus["OK"]:
                return originalStatus
            originalStatus = originalStatus["Value"]

            if currentStatus and currentStatus != originalStatus:
                return S_ERROR("Status changed in the DB: %s" % originalStatus)

            transIDAsDict = {transID: [originalStatus, transformationType]}
            dictOfProposedstatus = {transID: paramValue}
            # applying the state machine to the proposed status
            value = self._applyTransformationStatusStateMachine(transIDAsDict, dictOfProposedstatus, force)
        else:
            value = paramValue

        return rpcClient.setTransformationParameter(transID, paramName, value)

    def _applyTransformationStatusStateMachine(self, transIDAsDict, dictOfProposedstatus, force):
        """For easier extension, here we apply the state machine of the transformation status.
        VOs might want to replace the standard here with something they prefer.

        transIDAsDict is a dictionary with the transID as key and as value a list with [Status, Type]
        dictOfProposedstatus is a dictionary with the proposed status
        force is a boolean

        It returns the new status (the standard is just doing nothing: everything is possible)
        """
        return list(dictOfProposedstatus.values())[0]

    def isOK(self):
        return self.valid

    def addDirectory(self, path, force=False):
        rpcClient = self._getRPC()
        return rpcClient.addDirectory(path, force)



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-list-hosts
# Author :  Adrian Casajus
########################################################################
"""
List hosts

Example:
  $ dirac-admin-list-hosts
  dirac.in2p3.fr
  host-dirac.in2p3.fr
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("e", "extended", "Show extended info")

    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import exit as DIRACExit
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []
    extendedInfo = False

    for unprocSw in Script.getUnprocessedSwitches():
        if unprocSw[0] in ("e", "extended"):
            extendedInfo = True

    if not extendedInfo:
        result = diracAdmin.csListHosts()
        for host in result["Value"]:
            print(" %s" % host)
    else:
        result = diracAdmin.csDescribeHosts()
        print(diracAdmin.pPrint.pformat(result["Value"]))

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRACExit(exitCode)


if __name__ == "__main__":
    main()



""" VOBOXAvailabilityCommand module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# FIXME: NOT Usable ATM
# missing doNew, doCache, doMaster

from six.moves.urllib import parse as urlparse

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.ResourceStatusSystem.Command.Command import Command


class VOBOXAvailabilityCommand(Command):
    """
    Given an url pointing to a service on a vobox, use DIRAC ping against it.
    """

    def doCommand(self):
        """
        The Command pings a service on a vobox, it needs a service URL to ping it.

        :returns: a dict with the following:

          .. code-block:: python

            {
              'serviceUpTime' : <serviceUpTime>,
              'machineUpTime' : <machineUpTime>,
              'site'          : <site>,
              'system'        : <system>,
              'service'       : <service>
            }

        """

        # INPUT PARAMETERS

        if "serviceURL" not in self.args:
            return self.returnERROR(S_ERROR('"serviceURL" not found in self.args'))
        serviceURL = self.args["serviceURL"]

        ##

        parsed = urlparse.urlparse(serviceURL)
        site = parsed[1].split(":")[0]

        try:
            system, service = parsed[2].strip("/").split("/")
        except ValueError:
            return self.returnERROR(S_ERROR('"%s" seems to be a malformed url' % serviceURL))

        pinger = RPCClient(serviceURL)
        resPing = pinger.ping()

        if not resPing["OK"]:
            return self.returnERROR(resPing)

        serviceUpTime = resPing["Value"].get("service uptime", 0)
        machineUpTime = resPing["Value"].get("host uptime", 0)

        result = {
            "site": site,
            "system": system,
            "service": service,
            "serviceUpTime": serviceUpTime,
            "machineUpTime": machineUpTime,
        }

        return S_OK(result)

        # FIXME: how do we get the values !!


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



""" CAUpdateAgent is meant to be used in a multi-server installations
    where one server has some machinery of keeping up to date the CA's data
    and other servers are just synchronized with the master one without "official" CA installations locally.

    It's like installing CAs in the pilot in dirac-install but for the servers.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.FrameworkSystem.Client.BundleDeliveryClient import BundleDeliveryClient


class CAUpdateAgent(AgentModule):
    """just routinely calls BundleDeliveryClient.syncCAs()/syncCRLs()"""

    def initialize(self):
        self.am_setOption("PollingTime", 3600 * 6)
        return S_OK()

    def execute(self):
        """The main agent execution method"""
        bdc = BundleDeliveryClient()
        result = bdc.syncCAs()
        if not result["OK"]:
            self.log.error("Error while updating CAs", result["Message"])
        elif result["Value"]:
            self.log.info("CAs got updated")
        else:
            self.log.info("CAs are already synchronized")
        result = bdc.syncCRLs()
        if not result["OK"]:
            self.log.error("Error while updating CRLs", result["Message"])
        elif result["Value"]:
            self.log.info("CRLs got updated")
        else:
            self.log.info("CRLs are already synchronized")

        return S_OK()



############################################################################
#  GE class representing SGE batch system
#  10.11.2014
#  Author: A.T.
############################################################################

""" Torque.py is a DIRAC independent class representing Torque batch system.
    Torque objects are used as backend batch system representation for
    LocalComputingElement and SSHComputingElement classes

    The GE relies on the SubmitOptions parameter to choose the right queue.
    This should be specified in the Queue description in the CS. e.g.

    SubmitOption = -l ct=6000
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import re
import shlex
import subprocess
import os

__RCSID__ = "$Id$"


class GE(object):
    def submitJob(self, **kwargs):
        """Submit nJobs to the condor batch system"""
        resultDict = {}

        MANDATORY_PARAMETERS = ["Executable", "OutputDir", "ErrorDir", "SubmitOptions"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        nJobs = kwargs.get("NJobs", 1)
        preamble = kwargs.get("Preamble")

        outputs = []
        output = ""
        for _i in range(int(nJobs)):
            cmd = "%s; " % preamble if preamble else ""
            cmd += "qsub -o %(OutputDir)s -e %(ErrorDir)s -N DIRACPilot %(SubmitOptions)s %(Executable)s" % kwargs
            sp = subprocess.Popen(
                cmd,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            status = sp.returncode
            if status == 0:
                outputs.append(output)
            else:
                break

        if outputs:
            resultDict["Status"] = 0
            resultDict["Jobs"] = []
            for output in outputs:
                match = re.match(r"Your job (\d*) ", output)
                if match:
                    resultDict["Jobs"].append(match.groups()[0])
        else:
            resultDict["Status"] = status
            resultDict["Message"] = error

        return resultDict

    def killJob(self, **kwargs):
        """Kill jobs in the given list"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        jobIDList = kwargs.get("JobIDList")
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        successful = []
        failed = []
        errors = ""
        for job in jobIDList:
            sp = subprocess.Popen(
                shlex.split("qdel %s" % job),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            status = sp.returncode
            if status != 0:
                failed.append(job)
                errors += error
            else:
                successful.append(job)

        resultDict["Status"] = 0
        if failed:
            resultDict["Status"] = 1
            resultDict["Message"] = errors
        resultDict["Successful"] = successful
        resultDict["Failed"] = failed
        return resultDict

    def getJobStatus(self, **kwargs):
        """Get status of the jobs in the given list"""
        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        user = kwargs.get("User")
        if not user:
            user = os.environ.get("USER")
        if not user:
            resultDict["Status"] = -1
            resultDict["Message"] = "No user name"
            return resultDict
        jobIDList = kwargs.get("JobIDList")
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        sp = subprocess.Popen(
            shlex.split("qstat -u %s" % user),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode

        if status != 0:
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        jobDict = {}
        if output:
            lines = output.split("\n")
            for line in lines:
                l = line.strip()
                for job in jobIDList:
                    if l.startswith(job):
                        jobStatus = l.split()[4]
                        if jobStatus in ["Tt", "Tr"]:
                            jobDict[job] = "Done"
                        elif jobStatus in ["Rr", "r"]:
                            jobDict[job] = "Running"
                        elif jobStatus in ["qw", "h"]:
                            jobDict[job] = "Waiting"

        sp = subprocess.Popen(
            shlex.split("qstat -u %s -s -z" % user),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode

        if status == 0:
            if output:
                lines = output.split("\n")
                for line in lines:
                    l = line.strip()
                    for job in jobIDList:
                        if l.startswith(job):
                            jobDict[job] = "Done"

        if len(resultDict) != len(jobIDList):
            for job in jobIDList:
                if job not in jobDict:
                    jobDict[job] = "Unknown"

        # Final output
        status = 0
        resultDict["Status"] = 0
        resultDict["Jobs"] = jobDict
        return resultDict

    def getCEStatus(self, **kwargs):
        """Get the overall CE status"""
        resultDict = {}

        user = kwargs.get("User")
        if not user:
            user = os.environ.get("USER")
        if not user:
            resultDict["Status"] = -1
            resultDict["Message"] = "No user name"
            return resultDict

        cmd = "qstat -u %s" % user
        sp = subprocess.Popen(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode

        if status != 0:
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        waitingJobs = 0
        runningJobs = 0
        doneJobs = 0

        if output:
            lines = output.split("\n")
            for line in lines:
                if not line.strip():
                    continue
                if "DIRACPilot %s" % user in line:
                    jobStatus = line.split()[4]
                    if jobStatus in ["Tt", "Tr"]:
                        doneJobs += 1
                    elif jobStatus in ["Rr", "r"]:
                        runningJobs += 1
                    elif jobStatus in ["qw", "h"]:
                        waitingJobs = waitingJobs + 1

        # Final output
        resultDict["Status"] = 0
        resultDict["Waiting"] = waitingJobs
        resultDict["Running"] = runningJobs
        resultDict["Done"] = doneJobs
        return resultDict

    def getJobOutputFiles(self, **kwargs):
        """Get output file names and templates for the specific CE"""
        resultDict = {}
        MANDATORY_PARAMETERS = ["JobIDList", "OutputDir", "ErrorDir"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        outputDir = kwargs["OutputDir"]
        errorDir = kwargs["ErrorDir"]

        outputTemplate = "%s/DIRACPilot.o%%s" % outputDir
        errorTemplate = "%s/DIRACPilot.e%%s" % errorDir
        outputTemplate = os.path.expandvars(outputTemplate)
        errorTemplate = os.path.expandvars(errorTemplate)

        jobIDList = kwargs["JobIDList"]

        jobDict = {}
        for job in jobIDList:
            jobDict[job] = {}
            jobDict[job]["Output"] = outputTemplate % job
            jobDict[job]["Error"] = errorTemplate % job

        resultDict["Status"] = 0
        resultDict["Jobs"] = jobDict
        resultDict["OutputTemplate"] = outputTemplate
        resultDict["ErrorTemplate"] = errorTemplate

        return resultDict



#!/usr/bin/env python
"""
Get the list of all the user files.

Example:
  $ dirac-dms-user-lfns
  /formation/user/v/vhamar: 14 files, 6 sub-directories
  /formation/user/v/vhamar/newDir2: 0 files, 0 sub-directories
  /formation/user/v/vhamar/testDir: 0 files, 0 sub-directories
  /formation/user/v/vhamar/0: 0 files, 6 sub-directories
  /formation/user/v/vhamar/test: 0 files, 0 sub-directories
  /formation/user/v/vhamar/meta-test: 0 files, 0 sub-directories
  /formation/user/v/vhamar/1: 0 files, 4 sub-directories
  /formation/user/v/vhamar/0/994: 1 files, 0 sub-directories
  /formation/user/v/vhamar/0/20: 1 files, 0 sub-directories
  16 matched files have been put in formation-user-v-vhamar.lfns
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    days = 0
    months = 0
    years = 0
    wildcard = None
    baseDir = ""
    emptyDirsFlag = False
    Script.registerSwitch("D:", "Days=", "Match files older than number of days [%s]" % days)
    Script.registerSwitch("M:", "Months=", "Match files older than number of months [%s]" % months)
    Script.registerSwitch("Y:", "Years=", "Match files older than number of years [%s]" % years)
    Script.registerSwitch("w:", "Wildcard=", "Wildcard for matching filenames [All]")
    Script.registerSwitch("b:", "BaseDir=", "Base directory to begin search (default /[vo]/user/[initial]/[username])")
    Script.registerSwitch("e", "EmptyDirs", "Create a list of empty directories")

    Script.parseCommandLine(ignoreErrors=False)

    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "D" or switch[0].lower() == "days":
            days = int(switch[1])
        if switch[0] == "M" or switch[0].lower() == "months":
            months = int(switch[1])
        if switch[0] == "Y" or switch[0].lower() == "years":
            years = int(switch[1])
        if switch[0].lower() == "w" or switch[0].lower() == "wildcard":
            wildcard = "*" + switch[1]
        if switch[0].lower() == "b" or switch[0].lower() == "basedir":
            baseDir = switch[1]
        if switch[0].lower() == "e" or switch[0].lower() == "emptydirs":
            emptyDirsFlag = True

    import DIRAC
    from DIRAC import gLogger
    from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup
    from DIRAC.Core.Security.ProxyInfo import getProxyInfo
    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
    from datetime import datetime, timedelta
    import sys
    import os
    import time
    import fnmatch

    fc = FileCatalog()

    def isOlderThan(cTimeStruct, days):
        timeDelta = timedelta(days=days)
        maxCTime = datetime.utcnow() - timeDelta
        if cTimeStruct < maxCTime:
            return True
        return False

    withMetadata = False
    if days or months or years:
        withMetadata = True
    totalDays = 0
    if years:
        totalDays += 365 * years
    if months:
        totalDays += 30 * months
    if days:
        totalDays += days

    res = getProxyInfo(False, False)
    if not res["OK"]:
        gLogger.error("Failed to get client proxy information.", res["Message"])
        DIRAC.exit(2)
    proxyInfo = res["Value"]
    if proxyInfo["secondsLeft"] == 0:
        gLogger.error("Proxy expired")
        DIRAC.exit(2)
    username = proxyInfo["username"]
    vo = ""
    if "group" in proxyInfo:
        vo = getVOForGroup(proxyInfo["group"])
    if not baseDir:
        if not vo:
            gLogger.error("Could not determine VO")
            Script.showHelp()
        baseDir = "/%s/user/%s/%s" % (vo, username[0], username)

    baseDir = baseDir.rstrip("/")

    gLogger.notice("Will search for files in %s%s" % (baseDir, (" matching %s" % wildcard) if wildcard else ""))
    activeDirs = [baseDir]

    allFiles = []
    emptyDirs = []

    while len(activeDirs) > 0:
        currentDir = activeDirs.pop()
        res = fc.listDirectory(currentDir, withMetadata, timeout=360)
        if not res["OK"]:
            gLogger.error("Error retrieving directory contents", "%s %s" % (currentDir, res["Message"]))
        elif currentDir in res["Value"]["Failed"]:
            gLogger.error(
                "Error retrieving directory contents", "%s %s" % (currentDir, res["Value"]["Failed"][currentDir])
            )
        else:
            dirContents = res["Value"]["Successful"][currentDir]
            subdirs = dirContents["SubDirs"]
            files = dirContents["Files"]
            if not subdirs and not files:
                emptyDirs.append(currentDir)
                gLogger.notice("%s: empty directory" % currentDir)
            else:
                for subdir in sorted(subdirs, reverse=True):
                    if (not withMetadata) or isOlderThan(subdirs[subdir]["CreationDate"], totalDays):
                        activeDirs.append(subdir)
                for filename in sorted(files):
                    fileOK = False
                    if (not withMetadata) or isOlderThan(files[filename]["MetaData"]["CreationDate"], totalDays):
                        if wildcard is None or fnmatch.fnmatch(filename, wildcard):
                            fileOK = True
                    if not fileOK:
                        files.pop(filename)
                allFiles += sorted(files)

                if len(files) or len(subdirs):
                    gLogger.notice(
                        "%s: %d files%s, %d sub-directories"
                        % (currentDir, len(files), " matching" if withMetadata or wildcard else "", len(subdirs))
                    )

    outputFileName = "%s.lfns" % baseDir.replace("/%s" % vo, "%s" % vo).replace("/", "-")
    outputFile = open(outputFileName, "w")
    for lfn in sorted(allFiles):
        outputFile.write(lfn + "\n")
    outputFile.close()
    gLogger.notice("%d matched files have been put in %s" % (len(allFiles), outputFileName))

    if emptyDirsFlag:
        outputFileName = "%s.emptydirs" % baseDir.replace("/%s" % vo, "%s" % vo).replace("/", "-")
        outputFile = open(outputFileName, "w")
        for dir in sorted(emptyDirs):
            outputFile.write(dir + "\n")
        outputFile.close()
        gLogger.notice("%d empty directories have been put in %s" % (len(emptyDirs), outputFileName))

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" TokenAgent

  This agent inspect all elements, and resets their tokens if necessary.

The following options can be set for the TokenAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN TokenAgent
  :end-before: ##END
  :dedent: 2
  :caption: TokenAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from datetime import datetime, timedelta

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient

AGENT_NAME = "ResourceStatus/TokenAgent"


class TokenAgent(AgentModule):
    """
    TokenAgent is in charge of checking tokens assigned on resources.
    Notifications are sent to those users owning expiring tokens.
    """

    # Rss token
    __rssToken = "rs_svc"

    def __init__(self, *args, **kwargs):
        """c'tor"""

        AgentModule.__init__(self, *args, **kwargs)

        self.notifyHours = 12
        self.adminMail = ""

        self.rsClient = None
        self.tokenDict = {}
        self.diracAdmin = None

    def initialize(self):
        """TokenAgent initialization"""

        self.notifyHours = self.am_getOption("notifyHours", self.notifyHours)
        self.adminMail = self.am_getOption("adminMail", self.adminMail)

        self.rsClient = ResourceStatusClient()
        self.diracAdmin = DiracAdmin()

        return S_OK()

    def execute(self):
        """
        Looks for user tokens. If they are expired, or expiring, it notifies users.
        """

        # Initialized here, as it is needed empty at the beginning of the execution
        self.tokenDict = {}

        elements = ("Site", "Resource", "Node")

        for element in elements:

            self.log.info("Processing %s" % element)

            interestingTokens = self._getInterestingTokens(element)
            if not interestingTokens["OK"]:
                self.log.error(interestingTokens["Message"])
                continue
            interestingTokens = interestingTokens["Value"]

            processTokens = self._processTokens(element, interestingTokens)
            if not processTokens["OK"]:
                self.log.error(processTokens["Message"])
                continue

        notificationResult = self._notifyOfTokens()
        if not notificationResult["OK"]:
            self.log.error(notificationResult["Message"])

        return S_OK()

    def _getInterestingTokens(self, element):
        """
        Given an element, picks all the entries with TokenExpiration < now + X<hours>
        If the TokenOwner is not the rssToken ( rs_svc ), it is selected.
        """

        tokenExpLimit = datetime.utcnow() + timedelta(hours=self.notifyHours)

        tokenElements = self.rsClient.selectStatusElement(
            element, "Status", meta={"older": ["TokenExpiration", tokenExpLimit]}
        )

        if not tokenElements["OK"]:
            return tokenElements

        tokenColumns = tokenElements["Columns"]
        tokenElements = tokenElements["Value"]

        interestingTokens = []

        for tokenElement in tokenElements:

            tokenElement = dict(zip(tokenColumns, tokenElement))

            if tokenElement["TokenOwner"] != self.__rssToken:
                interestingTokens.append(tokenElement)

        return S_OK(interestingTokens)

    def _processTokens(self, element, tokenElements):
        """
        Given an element and a list of interesting token elements, updates the
        database if the token is expired, logs a message and adds
        """

        never = datetime.max

        for tokenElement in tokenElements:

            try:
                name = tokenElement["Name"]
                statusType = tokenElement["StatusType"]
                status = tokenElement["Status"]
                tokenOwner = tokenElement["TokenOwner"]
                tokenExpiration = tokenElement["TokenExpiration"]
            except KeyError as e:
                return S_ERROR(e)

            # If token has already expired
            if tokenExpiration < datetime.utcnow():
                _msg = '%s with statusType "%s" and owner %s EXPIRED'
                self.log.info(_msg % (name, statusType, tokenOwner))

                result = self.rsClient.addOrModifyStatusElement(
                    element,
                    "Status",
                    name=name,
                    statusType=statusType,
                    tokenOwner=self.__rssToken,
                    tokenExpiration=never,
                )
                if not result["OK"]:
                    return result

            else:
                _msg = '%s with statusType "%s" and owner %s -> %s'
                self.log.info(_msg % (name, statusType, tokenOwner, tokenExpiration))

            if tokenOwner not in self.tokenDict:
                self.tokenDict[tokenOwner] = []

            self.tokenDict[tokenOwner].append([tokenOwner, element, name, statusType, status, tokenExpiration])

        return S_OK()

    def _notifyOfTokens(self):
        """
        Splits interesing tokens between expired and expiring. Also splits them
        among users. It ends sending notifications to the users.
        """

        now = datetime.utcnow()

        adminExpired = []
        adminExpiring = []

        for tokenOwner, tokenLists in self.tokenDict.items():

            expired = []
            expiring = []

            for tokenList in tokenLists:

                if tokenList[5] < now:
                    expired.append(tokenList)
                    adminExpired.append(tokenList)
                else:
                    expiring.append(tokenList)
                    adminExpiring.append(tokenList)

            resNotify = self._notify(tokenOwner, expired, expiring)
            if not resNotify["OK"]:
                self.log.error("Failed to notify token owner", resNotify["Message"])

        if (adminExpired or adminExpiring) and self.adminMail:
            return self._notify(self.adminMail, adminExpired, adminExpiring)

        return S_OK()

    def _notify(self, tokenOwner, expired, expiring):
        """
        Given a token owner and a list of expired and expiring tokens, sends an
        email to the user.
        """

        subject = "RSS token summary for tokenOwner %s" % tokenOwner

        mail = "\nEXPIRED tokens ( RSS has taken control of them )\n"
        for tokenList in expired:

            mail += " ".join([str(x) for x in tokenList])
            mail += "\n"

        mail = "\nEXPIRING tokens ( RSS will take control of them )\n"
        for tokenList in expiring:

            mail += " ".join([str(x) for x in tokenList])
            mail += "\n"

        mail += "\n\n You can extend for another 24 hours using the web interface (Set token -> Acquire)\n"
        mail += " Or you can use the dirac-rss-set-token script\n\n"
        mail += "Through the same interfaces you can release the token any time\n"

        resEmail = self.diracAdmin.sendMail(tokenOwner, subject, mail)
        if not resEmail["OK"]:
            return S_ERROR('Cannot send email to user "%s"' % tokenOwner)

        return resEmail



########################################################################
# File :   ComputingElement.py
# Author : Stuart Paterson, A.T.
########################################################################

"""  The Computing Element class is a base class for all the various
     types CEs. It serves several purposes:

      - collects general CE related parameters to generate CE description
        for the job matching
      - provides logic for evaluation of the number of available CPU slots
      - provides logic for the proxy renewal while executing jobs

     The CE parameters are collected from the following sources, in hierarchy
     descending order:

      - parameters provided through setParameters() method of the class
      - parameters in /LocalSite configuration section
      - parameters in /LocalSite/<ceName>/ResourceDict configuration section
      - parameters in /LocalSite/ResourceDict configuration section
      - parameters in /LocalSite/<ceName> configuration section
      - parameters in /Resources/Computing/<ceName> configuration section
      - parameters in /Resources/Computing/CEDefaults configuration section

     The ComputingElement objects are usually instantiated with the help of
     ComputingElementFactory.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six
import os

from DIRAC import S_OK, S_ERROR, gLogger, version

from DIRAC.Core.Security import Properties
from DIRAC.Core.Security.VOMS import VOMS
from DIRAC.Core.Security.ProxyFile import writeToProxyFile
from DIRAC.Core.Security.ProxyInfo import getProxyInfoAsString
from DIRAC.Core.Security.ProxyInfo import formatProxyInfoAsString
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Utilities.Time import dateTime, second
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Config import gConfig
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.WorkloadManagementSystem.Utilities.JobParameters import (
    getNumberOfProcessors,
    getNumberOfGPUs,
)

INTEGER_PARAMETERS = ["CPUTime", "NumberOfProcessors", "NumberOfPayloadProcessors", "MaxRAM"]
FLOAT_PARAMETERS = ["WaitingToRunningRatio"]

__RCSID__ = "$Id$"

LIST_PARAMETERS = ["Tag", "RequiredTag"]
WAITING_TO_RUNNING_RATIO = 0.5
MAX_WAITING_JOBS = 1
MAX_TOTAL_JOBS = 1


class ComputingElement(object):
    """ComputingElement base class"""

    #############################################################################

    def __init__(self, ceName):
        """Standard constructor"""
        self.log = gLogger.getSubLogger(ceName)
        self.ceName = ceName
        self.ceType = ""
        self.ceParameters = {}
        self.proxy = ""
        self.valid = None
        self.mandatoryParameters = []
        self.batchSystem = None
        self.taskResults = {}
        self.minProxyTime = gConfig.getValue("/Registry/MinProxyLifeTime", 10800)  # secs
        self.defaultProxyTime = gConfig.getValue("/Registry/DefaultProxyLifeTime", 43200)  # secs
        self.proxyCheckPeriod = gConfig.getValue("/Registry/ProxyCheckingPeriod", 3600)  # secs

        self.initializeParameters()
        self.log.debug("CE parameters", self.ceParameters)

    def setProxy(self, proxy, valid=0):
        """Set proxy for this instance"""
        self.proxy = proxy
        self.valid = dateTime() + second * valid
        return S_OK()

    def _prepareProxy(self):
        """Set the environment variable X509_USER_PROXY"""
        if not self.proxy:
            result = getProxyInfo()
            if not result["OK"]:
                return S_ERROR("No proxy available")
            if "path" in result["Value"]:
                os.environ["X509_USER_PROXY"] = result["Value"]["path"]
                return S_OK()
        else:
            result = gProxyManager.dumpProxyToFile(self.proxy, requiredTimeLeft=self.minProxyTime)
            if not result["OK"]:
                return result
            os.environ["X509_USER_PROXY"] = result["Value"]

        self.log.debug("Set proxy variable X509_USER_PROXY to %s" % os.environ["X509_USER_PROXY"])
        return S_OK()

    def isProxyValid(self, valid=1000):
        """Check if the stored proxy is valid"""
        if not self.valid:
            result = S_ERROR("Proxy is not valid for the requested length")
            result["Value"] = 0
            return result
        delta = self.valid - dateTime()
        totalSeconds = delta.days * 86400 + delta.seconds
        if totalSeconds > valid:
            return S_OK(totalSeconds - valid)

        result = S_ERROR("Proxy is not valid for the requested length")
        result["Value"] = totalSeconds - valid
        return result

    def initializeParameters(self):
        """Initialize the CE parameters after they are collected from various sources"""

        self.log.debug("Initializing the CE parameters")

        # Collect global defaults first
        for section in ["/Resources/Computing/CEDefaults", "/Resources/Computing/%s" % self.ceName]:
            result = gConfig.getOptionsDict(section)

            self.log.debug(result)

            if result["OK"]:
                ceOptions = result["Value"]
                for key in ceOptions:
                    if key in INTEGER_PARAMETERS:
                        ceOptions[key] = int(ceOptions[key])
                    if key in FLOAT_PARAMETERS:
                        ceOptions[key] = float(ceOptions[key])
                    if key in LIST_PARAMETERS:
                        ceOptions[key] = gConfig.getValue(os.path.join(section, key), [])
                self.ceParameters.update(ceOptions)

        # Get local CE configuration
        localConfigDict = getCEConfigDict(self.ceName)
        self.ceParameters.update(localConfigDict)

        # Adds site level parameters
        section = "/LocalSite"
        result = gConfig.getOptionsDict(section)
        if result["OK"] and result["Value"]:
            localSiteParameters = result["Value"]
            self.log.debug("Local site parameters are: %s" % (localSiteParameters))
            for option, value in localSiteParameters.items():
                if option == "Architecture":
                    self.ceParameters["Platform"] = value
                    self.ceParameters["Architecture"] = value
                elif option == "LocalSE":
                    self.ceParameters["LocalSE"] = value.split(", ")
                else:
                    self.ceParameters[option] = value

        self._addCEConfigDefaults()

    def isValid(self):
        """Check the sanity of the Computing Element definition"""
        for par in self.mandatoryParameters:
            if par not in self.ceParameters:
                return S_ERROR("Missing Mandatory Parameter in Configuration: %s" % par)
        return S_OK()

    #############################################################################
    def _addCEConfigDefaults(self):
        """Method to make sure all necessary Configuration Parameters are defined"""
        self.ceParameters["WaitingToRunningRatio"] = float(
            self.ceParameters.get("WaitingToRunningRatio", WAITING_TO_RUNNING_RATIO)
        )
        self.ceParameters["MaxWaitingJobs"] = int(self.ceParameters.get("MaxWaitingJobs", MAX_WAITING_JOBS))
        self.ceParameters["MaxTotalJobs"] = int(self.ceParameters.get("MaxTotalJobs", MAX_TOTAL_JOBS))

    def _reset(self):
        """Make specific CE parameter adjustments after they are collected or added"""
        return S_OK()

    def loadBatchSystem(self, batchSystemName):
        """Instantiate object representing the backend batch system

        :param str batchSystemName: name of the batch system
        """
        if batchSystemName is None:
            batchSystemName = self.ceParameters["BatchSystem"]

        objectLoader = ObjectLoader()
        result = objectLoader.loadObject("Resources.Computing.BatchSystems.%s" % batchSystemName, batchSystemName)
        if not result["OK"]:
            self.log.error("Failed to load batch object: %s" % result["Message"])
            return result
        batchClass = result["Value"]
        batchModuleFile = result["ModuleFile"]
        self.batchSystem = batchClass()
        self.log.info("Batch system class from module: ", batchModuleFile)
        return S_OK()

    def setParameters(self, ceOptions):
        """Add parameters from the given dictionary overriding the previous values

        :param dict ceOptions: CE parameters dictionary to update already defined ones
        """
        self.ceParameters.update(ceOptions)

        # At this point we can know the exact type of CE,
        # try to get generic parameters for this type
        ceType = self.ceParameters.get("CEType")
        if ceType:
            result = gConfig.getOptionsDict("/Resources/Computing/%s" % ceType)
            if result["OK"]:
                generalCEDict = result["Value"]
                generalCEDict.update(self.ceParameters)
                self.ceParameters = generalCEDict

        # If NumberOfProcessors/GPUs is present in the description but is equal to zero
        # interpret it as needing local evaluation
        if self.ceParameters.get("NumberOfProcessors", -1) == 0:
            self.ceParameters["NumberOfProcessors"] = getNumberOfProcessors()
        if self.ceParameters.get("NumberOfGPUs", -1) == 0:
            self.ceParameters["NumberOfGPUs"] = getNumberOfGPUs()

        for key in ceOptions:
            if key in INTEGER_PARAMETERS:
                self.ceParameters[key] = int(self.ceParameters[key])
            if key in FLOAT_PARAMETERS:
                self.ceParameters[key] = float(self.ceParameters[key])

        return self._reset()

    #############################################################################
    def setCPUTimeLeft(self, cpuTimeLeft=None):
        """Update the CPUTime parameter of the CE classAd, necessary for running in filling mode"""
        if not cpuTimeLeft:
            # do nothing
            return S_OK()
        try:
            intCPUTimeLeft = int(cpuTimeLeft)
            self.ceParameters["CPUTime"] = intCPUTimeLeft
            return S_OK(intCPUTimeLeft)
        except ValueError:
            return S_ERROR("Wrong type for setCPUTimeLeft argument")

    #############################################################################
    def available(self, jobIDList=None):
        """This method returns the number of available slots in the target CE. The CE
        instance polls for waiting and running jobs and compares to the limits
        in the CE parameters.

        :param list jobIDList: list of already existing job IDs to be checked against
        """

        # If there are no already registered jobs
        if jobIDList is not None and not jobIDList:
            result = S_OK()
            result["RunningJobs"] = 0
            result["WaitingJobs"] = 0
            result["SubmittedJobs"] = 0
        else:
            result = self.ceParameters.get("CEType")
            if result and result == "CREAM":
                result = self.getCEStatus(jobIDList)  # pylint: disable=too-many-function-args
            else:
                result = self.getCEStatus()
            if not result["OK"]:
                return result
        runningJobs = result["RunningJobs"]
        waitingJobs = result["WaitingJobs"]
        submittedJobs = result["SubmittedJobs"]
        availableProcessors = result.get("AvailableProcessors")
        ceInfoDict = dict(result)

        maxTotalJobs = int(self.ceParameters.get("MaxTotalJobs", 0))
        ceInfoDict["MaxTotalJobs"] = maxTotalJobs
        waitingToRunningRatio = float(self.ceParameters.get("WaitingToRunningRatio", 0.0))
        # if there are no Running job we can submit to get at most 'MaxWaitingJobs'
        # if there are Running jobs we can increase this to get a ratio W / R 'WaitingToRunningRatio'
        maxWaitingJobs = int(max(int(self.ceParameters.get("MaxWaitingJobs", 0)), runningJobs * waitingToRunningRatio))

        self.log.verbose("Max Number of Jobs:", maxTotalJobs)
        self.log.verbose("Max W/R Ratio:", waitingToRunningRatio)
        self.log.verbose("Max Waiting Jobs:", maxWaitingJobs)

        # Determine how many more jobs can be submitted
        message = "%s CE: SubmittedJobs=%s" % (self.ceName, submittedJobs)
        message += ", WaitingJobs=%s, RunningJobs=%s" % (waitingJobs, runningJobs)
        totalJobs = runningJobs + waitingJobs

        message += ", MaxTotalJobs=%s" % (maxTotalJobs)

        if totalJobs >= maxTotalJobs:
            self.log.verbose("Max Number of Jobs reached:", maxTotalJobs)
            result["Value"] = 0
            message = "There are %s waiting jobs and total jobs %s >= %s max total jobs" % (
                waitingJobs,
                totalJobs,
                maxTotalJobs,
            )
        else:
            additionalJobs = 0
            if waitingJobs < maxWaitingJobs:
                additionalJobs = maxWaitingJobs - waitingJobs
                if totalJobs + additionalJobs >= maxTotalJobs:
                    additionalJobs = maxTotalJobs - totalJobs
            # For SSH CE case
            if int(self.ceParameters.get("MaxWaitingJobs", 0)) == 0:
                additionalJobs = maxTotalJobs - runningJobs

            if availableProcessors is not None:
                additionalJobs = min(additionalJobs, availableProcessors)
            result["Value"] = additionalJobs

        result["Message"] = message
        result["CEInfoDict"] = ceInfoDict
        return result

    #############################################################################
    def writeProxyToFile(self, proxy):
        """CE helper function to write a CE proxy string to a file."""
        result = writeToProxyFile(proxy)
        if not result["OK"]:
            self.log.error("Could not write proxy to file", result["Message"])
            return result

        proxyLocation = result["Value"]
        result = getProxyInfoAsString(proxyLocation)
        if not result["OK"]:
            self.log.error("Could not get proxy info", result)
            return result
        else:
            self.log.info("Payload proxy information:")
            print(result["Value"])

        return S_OK(proxyLocation)

    #############################################################################
    def _monitorProxy(self, payloadProxy=None):
        """Base class for the monitor and update of the payload proxy, to be used in
        derived classes for the basic renewal of the proxy, if further actions are
        necessary they should be implemented there

        :param str payloadProxy: location of the payload proxy file

        :returns: S_OK(filename)/S_ERROR
        """
        if not payloadProxy:
            return S_ERROR("No payload proxy")

        # This will get the pilot proxy
        ret = getProxyInfo()
        if not ret["OK"]:
            pilotProxy = None
        else:
            pilotProxy = ret["Value"]["path"]
            self.log.notice("Pilot Proxy:", pilotProxy)

        retVal = getProxyInfo(payloadProxy)
        if not retVal["OK"]:
            self.log.error("Could not get payload proxy info", retVal)
            return retVal
        self.log.verbose("Payload Proxy information:\n%s" % formatProxyInfoAsString(retVal["Value"]))

        payloadProxyDict = retVal["Value"]
        payloadSecs = payloadProxyDict["chain"].getRemainingSecs()["Value"]
        if payloadSecs > self.minProxyTime:
            self.log.verbose("No need to renew payload Proxy")
            return S_OK()

        # if there is no pilot proxy, assume there is a certificate and try a renewal
        if not pilotProxy:
            self.log.info("Using default credentials to get a new payload Proxy")
            return gProxyManager.renewProxy(
                proxyToBeRenewed=payloadProxy,
                minLifeTime=self.minProxyTime,
                newProxyLifeTime=self.defaultProxyTime,
                proxyToConnect=pilotProxy,
            )

        # if there is pilot proxy
        retVal = getProxyInfo(pilotProxy)
        if not retVal["OK"]:
            return retVal
        pilotProxyDict = retVal["Value"]

        if "groupProperties" not in pilotProxyDict:
            self.log.error("Invalid Pilot Proxy", "Group has no properties defined")
            return S_ERROR("Proxy has no group properties defined")

        pilotProps = pilotProxyDict["groupProperties"]

        # if running with a pilot proxy, use it to renew the proxy of the payload
        if Properties.PILOT in pilotProps or Properties.GENERIC_PILOT in pilotProps:
            self.log.info("Using Pilot credentials to get a new payload Proxy")
            return gProxyManager.renewProxy(
                proxyToBeRenewed=payloadProxy,
                minLifeTime=self.minProxyTime,
                newProxyLifeTime=self.defaultProxyTime,
                proxyToConnect=pilotProxy,
            )

        # if we are running with other type of proxy check if they are for the same user and group
        # and copy the pilot proxy if necessary

        self.log.info("Trying to copy pilot Proxy to get a new payload Proxy")
        pilotProxySecs = pilotProxyDict["chain"].getRemainingSecs()["Value"]
        if pilotProxySecs <= payloadSecs:
            errorStr = "Pilot Proxy is not longer than payload Proxy"
            self.log.error(errorStr)
            return S_ERROR("Can not renew by copy: %s" % errorStr)

        # check if both proxies belong to the same user and group
        pilotDN = pilotProxyDict["chain"].getIssuerCert()["Value"].getSubjectDN()["Value"]
        retVal = pilotProxyDict["chain"].getDIRACGroup()
        if not retVal["OK"]:
            return retVal
        pilotGroup = retVal["Value"]

        payloadDN = payloadProxyDict["chain"].getIssuerCert()["Value"].getSubjectDN()["Value"]
        retVal = payloadProxyDict["chain"].getDIRACGroup()
        if not retVal["OK"]:
            return retVal
        payloadGroup = retVal["Value"]
        if pilotDN != payloadDN or pilotGroup != payloadGroup:
            errorStr = "Pilot Proxy and payload Proxy do not have same DN and Group"
            self.log.error(errorStr)
            return S_ERROR("Can not renew by copy: %s" % errorStr)

        if pilotProxyDict.get("hasVOMS", False):
            return pilotProxyDict["chain"].dumpAllToFile(payloadProxy)

        attribute = Registry.getVOMSAttributeForGroup(payloadGroup)
        vo = Registry.getVOMSVOForGroup(payloadGroup)

        retVal = VOMS().setVOMSAttributes(pilotProxyDict["chain"], attribute=attribute, vo=vo)
        if not retVal["OK"]:
            return retVal

        chain = retVal["Value"]
        return chain.dumpAllToFile(payloadProxy)

    def getDescription(self):
        """Get CE description as a dictionary.

        This is called by the JobAgent for the case of "inner" CEs.
        """

        ceDict = {}
        for option, value in self.ceParameters.items():
            if isinstance(value, list):
                ceDict[option] = value
            elif isinstance(value, six.string_types):
                try:
                    ceDict[option] = int(value)
                except ValueError:
                    ceDict[option] = value
            elif isinstance(value, six.integer_types + (float,)):
                ceDict[option] = value
            else:
                self.log.warn("Type of option %s = %s not determined" % (option, value))

        release = gConfig.getValue("/LocalSite/ReleaseVersion", version)
        ceDict["DIRACVersion"] = release
        ceDict["ReleaseVersion"] = release
        project = gConfig.getValue("/LocalSite/ReleaseProject", "")
        if project:
            ceDict["ReleaseProject"] = project

        # the getCEStatus is implemented in each of the specific CE classes
        result = self.getCEStatus()
        if result["OK"]:
            ceDict["NumberOfProcessors"] = result.get("AvailableProcessors", result.get("NumberOfProcessors", 1))
        else:
            self.log.error(
                "Failure getting CE status", "(we keep going without the number of waiting and running pilots/jobs)"
            )

        return S_OK(ceDict)

    #############################################################################
    def sendOutput(self, stdid, line):  # pylint: disable=unused-argument, no-self-use
        """Callback function such that the results from the CE may be returned."""
        print(line)

    #############################################################################
    def submitJob(self, executableFile, proxy, **kwargs):  # pylint: disable=unused-argument
        """Method to submit job, should be overridden in sub-class."""
        name = "submitJob()"
        self.log.error("ComputingElement should be implemented in a subclass", name)
        return S_ERROR("ComputingElement: %s should be implemented in a subclass" % (name))

    #############################################################################
    def getCEStatus(self):
        """Method to get dynamic job information, can be overridden in sub-class."""
        name = "getCEStatus()"
        self.log.error("ComputingElement should be implemented in a subclass", name)
        return S_ERROR("ComputingElement: %s should be implemented in a subclass" % (name))

    #############################################################################
    def shutdown(self):
        """Optional method to shutdown the (Inner) Computing Element"""
        return S_OK(self.taskResults)


def getCEConfigDict(ceName):
    """Look into LocalSite for configuration Parameters for this CE"""
    ceConfigDict = {}
    if ceName:
        result = gConfig.getOptionsDict("/LocalSite/%s" % ceName)
        if result["OK"]:
            ceConfigDict = result["Value"]
    return ceConfigDict



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-get-wn
# Author :  Philippe Charpentier
########################################################################
"""
Get WNs for a selection of jobs
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime
from functools import cmp_to_key

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    site = "BOINC.World.org"
    status = ["Running"]
    minorStatus = None
    workerNodes = None
    since = None
    date = "today"
    full = False
    until = None
    batchIDs = None
    Script.registerSwitch("", "Site=", "   Select site (default: %s)" % site)
    Script.registerSwitch("", "Status=", "   Select status (default: %s)" % status)
    Script.registerSwitch("", "MinorStatus=", "   Select minor status")
    Script.registerSwitch("", "WorkerNode=", "  Select WN")
    Script.registerSwitch("", "BatchID=", "  Select batch jobID")
    Script.registerSwitch("", "Since=", "   Date since when to select jobs, or number of days (default: today)")
    Script.registerSwitch("", "Date=", "   Specify the date (check for a full day)")
    Script.registerSwitch("", "Full", "   Printout full list of job (default: False except if --WorkerNode)")

    Script.parseCommandLine()
    from DIRAC import gLogger
    from DIRAC.Interfaces.API.Dirac import Dirac
    from DIRAC.WorkloadManagementSystem.Client.JobMonitoringClient import JobMonitoringClient

    switches = Script.getUnprocessedSwitches()
    for switch in switches:
        if switch[0] == "Site":
            site = switch[1]
        elif switch[0] == "MinorStatus":
            minorStatus = switch[1]
        elif switch[0] == "Status":
            if switch[1].lower() == "all":
                status = [None]
            else:
                status = switch[1].split(",")
        elif switch[0] == "WorkerNode":
            workerNodes = switch[1].split(",")
        elif switch[0] == "BatchID":
            try:
                batchIDs = [int(id) for id in switch[1].split(",")]
            except Exception:
                gLogger.error("Invalid jobID", switch[1])
                DIRAC.exit(1)
        elif switch[0] == "Full":
            full = True
        elif switch[0] == "Date":
            since = switch[1].split()[0]
            until = str(datetime.datetime.strptime(since, "%Y-%m-%d") + datetime.timedelta(days=1)).split()[0]
        elif switch[0] == "Since":
            date = switch[1].lower()
            if date == "today":
                since = None
            elif date == "yesterday":
                since = 1
            elif date == "ever":
                since = 2 * 365
            elif date.isdigit():
                since = int(date)
                date += " days"
            else:
                since = date
            if isinstance(since, int):
                since = str(datetime.datetime.now() - datetime.timedelta(days=since)).split()[0]

    if workerNodes or batchIDs:
        # status = [None]
        full = True

    monitoring = JobMonitoringClient()
    dirac = Dirac()

    # Get jobs according to selection
    jobs = set()
    for stat in status:
        res = dirac.selectJobs(site=site, date=since, status=stat, minorStatus=minorStatus)
        if not res["OK"]:
            gLogger.error("Error selecting jobs", res["Message"])
            DIRAC.exit(1)
        allJobs = set(int(job) for job in res["Value"])
        if until:
            res = dirac.selectJobs(site=site, date=until, status=stat)
            if not res["OK"]:
                gLogger.error("Error selecting jobs", res["Message"])
                DIRAC.exit(1)
            allJobs -= set(int(job) for job in res["Value"])
        jobs.update(allJobs)
    if not jobs:
        gLogger.always("No jobs found...")
        DIRAC.exit(0)
    # res = monitoring.getJobsSummary( jobs )
    # print eval( res['Value'] )[jobs[0]]

    allJobs = set()
    result = {}
    wnJobs = {}
    gLogger.always("%d jobs found" % len(jobs))
    # Get host name
    for job in jobs:
        res = monitoring.getJobParameter(job, "HostName")
        node = res.get("Value", {}).get("HostName", "Unknown")
        res = monitoring.getJobParameter(job, "LocalJobID")
        batchID = res.get("Value", {}).get("LocalJobID", "Unknown")
        if workerNodes:
            if not [wn for wn in workerNodes if node.startswith(wn)]:
                continue
            allJobs.add(job)
        if batchIDs:
            if batchID not in batchIDs:
                continue
            allJobs.add(job)
        if full or status == [None]:
            allJobs.add(job)
        result.setdefault(job, {})["Status"] = status
        result[job]["Node"] = node
        result[job]["LocalJobID"] = batchID
        wnJobs[node] = wnJobs.setdefault(node, 0) + 1

    # If necessary get jobs' status
    statusCounters = {}
    if allJobs:
        allJobs = sorted(allJobs, reverse=True)
        res = monitoring.getJobsStates(allJobs)
        if not res["OK"]:
            gLogger.error("Error getting job parameter", res["Message"])
        else:
            jobStates = res["Value"]
            for job in allJobs:
                stat = (
                    jobStates.get(job, {}).get("Status", "Unknown")
                    + "; "
                    + jobStates.get(job, {}).get("MinorStatus", "Unknown")
                    + "; "
                    + jobStates.get(job, {}).get("ApplicationStatus", "Unknown")
                )
                result[job]["Status"] = stat
                statusCounters[stat] = statusCounters.setdefault(stat, 0) + 1
    elif not workerNodes and not batchIDs:
        allJobs = sorted(jobs, reverse=True)

    # Print out result
    if workerNodes or batchIDs:
        gLogger.always("Found %d jobs at %s, WN %s (since %s):" % (len(allJobs), site, workerNodes, date))
        if allJobs:
            gLogger.always("List of jobs:", ",".join([str(job) for job in allJobs]))
    else:
        if status == [None]:
            gLogger.always("Found %d jobs at %s (since %s):" % (len(allJobs), site, date))
            for stat in sorted(statusCounters):
                gLogger.always("%d jobs %s" % (statusCounters[stat], stat))
        else:
            gLogger.always("Found %d jobs %s at %s (since %s):" % (len(allJobs), status, site, date))
        gLogger.always(
            "List of WNs:",
            ",".join(
                [
                    "%s (%d)" % (node, wnJobs[node])
                    for node in sorted(wnJobs, key=cmp_to_key(lambda n1, n2: (wnJobs[n2] - wnJobs[n1])))
                ]
            ),
        )
    if full:
        if workerNodes or batchIDs:
            nodeJobs = {}
            for job in allJobs:
                status = result[job]["Status"]
                node = result[job]["Node"].split(".")[0]
                jobID = result[job].get("LocalJobID")
                nodeJobs.setdefault(node, []).append((jobID, job, status))
            if not workerNodes:
                workerNodes = sorted(nodeJobs)
            for node in workerNodes:
                for job in nodeJobs.get(node.split(".")[0], []):
                    gLogger.always("%s " % node + "(%s): %s - %s" % job)
        else:
            for job in allJobs:
                status = result[job]["Status"]
                node = result[job]["Node"]
                jobID = result[job].get("LocalJobID")
                gLogger.always("%s (%s): %s - %s" % (node, jobID, job, status))


if __name__ == "__main__":
    main()



""" RocciEndpoint class is the implementation of the rocci interface to
    a cloud endpoint via rOCCI-cli
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os
import json
import subprocess
import tempfile

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.File import makeGuid
from DIRAC.Resources.Cloud.Endpoint import Endpoint

__RCSID__ = "$Id$"


class RocciEndpoint(Endpoint):
    def __init__(self, parameters=None):
        super(RocciEndpoint, self).__init__(parameters=parameters)
        # logger
        self.log = gLogger.getSubLogger("RocciEndpoint")
        self.valid = False
        result = self.initialize()
        if result["OK"]:
            self.log.debug("RocciEndpoint created and validated")
            self.valid = True
        else:
            self.log.error(result["Message"])

    def initialize(self):

        availableParams = {
            "EndpointUrl": "endpoint",
            "Timeout": "timeout",
            "Auth": "auth",
            "User": "username",
            "Password": "password",
            "UserCred": "user-cred",
            "VOMS": "voms",
        }

        self.__occiBaseCmd = ["occi", "--skip-ca-check", "--output-format", "json_extended"]
        for var in availableParams:
            if var in self.parameters:
                self.__occiBaseCmd += ["--%s" % availableParams[var], "%s" % self.parameters[var]]

        result = self.__checkConnection()
        return result

    def __filterCommand(self, cmd):
        filteredCmd = []
        mask = False
        for arg in cmd:
            if mask:
                filteredCmd.append("xxxxxx")
                mask = False
            else:
                filteredCmd.append(arg)

            if arg in ["--username", "--password"]:
                mask = True
        return " ".join(filteredCmd)

    def __occiCommand(self, actionArgs):
        try:
            finalCmd = self.__occiBaseCmd + actionArgs
            self.log.debug("Running command:", self.__filterCommand(finalCmd))
            p = subprocess.Popen(finalCmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = p.communicate()
            if p.returncode != 0:
                return S_ERROR("occi command exit with error %s: %s" % (p.returncode, stderr))
        except Exception as e:
            return S_ERROR("Can not run occi command")

        return S_OK(stdout)

    def __checkConnection(self):
        """
        Checks connection status by trying to list the images.

        :return: S_OK | S_ERROR
        """
        actionArgs = ["--action", "list", "--resource", "os_tpl"]
        result = self.__occiCommand(actionArgs)
        if not result["OK"]:
            return result

        return S_OK()

    def __getImageByName(self, imageName):
        """
        Given the imageName, returns the current image object from the server.

        :Parameters:
          **imageName** - `string`

        :return: S_OK( image ) | S_ERROR
        """
        # the libcloud library, throws Exception. Nothing to do.
        actionArgs = ["--action", "describe", "--resource", "os_tpl"]
        result = self.__occiCommand(actionArgs)
        if not result["OK"]:
            return result

        imageIds = []
        for image in json.loads(result["Value"]):
            if image["title"] == imageName:
                imageIds.append(image["term"])

        if not imageIds:
            return S_ERROR("Image %s not found" % imageName)

        if len(imageIds) > 1:
            self.log.warn("More than one image found", '%s images with name "%s"' % (len(imageIds), imageName))

        return S_OK(imageIds[-1])

    def createInstances(self, vmsToSubmit):
        outputDict = {}

        for nvm in range(vmsToSubmit):
            instanceID = makeGuid()[:8]
            result = self.createInstance(instanceID)
            if result["OK"]:
                occiId, nodeDict = result["Value"]
                self.log.debug("Created VM instance %s/%s" % (occiId, instanceID))
                outputDict[occiId] = nodeDict
            else:
                self.log.error("Create Rocci instance error:", result["Message"])
                break

        return S_OK(outputDict)

    def createInstance(self, instanceID=""):
        if not instanceID:
            instanceID = makeGuid()[:8]

        self.parameters["VMUUID"] = instanceID
        self.parameters["VMType"] = self.parameters.get("CEType", "EC2")

        actionArgs = ["--action", "create"]
        actionArgs += ["--resource", "compute"]

        # Image
        if "ImageID" in self.parameters and "ImageName" not in self.parameters:
            result = self.__getImageByName(self.parameters["ImageName"])
            if not result["OK"]:
                return result
            imageId = result["Value"]
        elif "ImageID" in self.parameters:
            result = self.__occiCommand(
                ["--action", "describe", "--resource", "os_tpl#%s" % self.parameters["ImageID"]]
            )
            if not result["OK"]:
                return S_ERROR("Failed to get image for ID %s" % self.parameters["ImageID"], result["Message"])
            imageId = self.parameters["ImageID"]
        else:
            return S_ERROR("No image specified")
        actionArgs += ["--mixin", "os_tpl#%s" % imageId]

        # Optional flavor name
        if "FlavorName" in self.parameters:
            result = self.__occiCommand(
                ["--action", "describe", "--resource", "resource_tpl#%s" % self.parameters["FlavorName"]]
            )
            if not result["OK"]:
                return S_ERROR("Failed to get flavor %s" % self.parameters["FlavorName"], result["Message"])
            actionArgs += ["--mixin", "resource_tpl#%s" % self.parameters["FlavorName"]]

        # Instance name
        actionArgs += ["--attribute", "occi.core.title=DIRAC_%s" % instanceID]

        # Other params
        for param in []:
            if param in self.parameters:
                actionArgs += ["--%s" % param, "%s" % self.parameters[param]]

        self.log.info("Creating node:")
        self.log.verbose(" ".join(actionArgs))

        # User data
        result = self._createUserDataScript()
        if not result["OK"]:
            return result
        #    actionArgs += ['--context', 'user_data=%s' % str( result['Value'] )]
        f = tempfile.NamedTemporaryFile(delete=False)
        f.write(str(result["Value"]))
        f.close()
        self.log.debug("Write user_data to temp file:", f.name)
        actionArgs += ["--context", "user_data=file://%s" % f.name]

        # Create the VM instance now
        result = self.__occiCommand(actionArgs)
        os.unlink(f.name)
        if not result["OK"]:
            errmsg = "Error in rOCCI create instances: %s" % result["Message"]
            self.log.error(errmsg)
            return S_ERROR(errmsg)

        occiId = result["Value"].strip()

        # Properties of the instance
        nodeDict = {}
        nodeDict["InstanceID"] = instanceID
        result = self.__occiCommand(["--action", "describe", "--resource", occiId])
        if result["OK"]:
            nodeInfo = json.loads(result["Value"])
            try:
                nodeDict["NumberOfProcessors"] = nodeInfo[0]["attributes"]["occi"]["compute"]["cores"]
                nodeDict["RAM"] = nodeInfo[0]["attributes"]["occi"]["compute"]["memory"]
            except Exception as e:
                nodeDict["NumberOfProcessors"] = 1
        else:
            nodeDict["NumberOfProcessors"] = 1

        return S_OK((occiId, nodeDict))

    def stopVM(self, nodeID, publicIP=""):
        actionArgs = ["--action", "delete", "--resource", nodeID]
        result = self.__occiCommand(actionArgs)
        if not result["OK"]:
            errmsg = "Can not terminate instance %s: %s" % (nodeID, result["Message"])
            self.log.error(errmsg)
            return S_ERROR(errmsg)

        return S_OK()



#!/usr/bin/env python
########################################################################
# $Header:  $
########################################################################
"""
Remove the given file or a list of files from the File Catalog

Example:
  $ dirac-dms-remove-catalog-files   /formation/user/v/vhamar/1/1134/StdOut
  Successfully removed 1 catalog files.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC import exit as dexit
from DIRAC import gLogger


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(("LocalFile: Path to local file containing LFNs", "LFN:       Logical File Names"))
    Script.parseCommandLine()

    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

    allowUsers = Operations().getValue("DataManagement/AllowUserReplicaManagement", False)

    from DIRAC.Core.Security.ProxyInfo import getProxyInfo

    res = getProxyInfo()
    if not res["OK"]:
        gLogger.fatal("Can't get proxy info", res["Message"])
        dexit(1)
    properties = res["Value"].get("groupProperties", [])

    if not allowUsers:
        if "FileCatalogManagement" not in properties:
            gLogger.error("You need to use a proxy from a group with FileCatalogManagement")
            dexit(5)

    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

    fc = FileCatalog()
    import os

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    args = Script.getPositionalArgs()

    inputFileName = args[0]

    if os.path.exists(inputFileName):
        inputFile = open(inputFileName, "r")
        string = inputFile.read()
        lfns = [lfn.strip() for lfn in string.splitlines()]
        inputFile.close()
    else:
        lfns = [inputFileName]

    res = fc.removeFile(lfns)
    if not res["OK"]:
        print("Error:", res["Message"])
        dexit(1)
    for lfn in sorted(res["Value"]["Failed"].keys()):
        message = res["Value"]["Failed"][lfn]
        print("Error: failed to remove %s: %s" % (lfn, message))
    print("Successfully removed %d catalog files." % (len(res["Value"]["Successful"])))


if __name__ == "__main__":
    main()



""" FileCatalogFactory class to create file catalog client objects according to the
    configuration description
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getCatalogPath
from DIRAC.Resources.Catalog.FileCatalogProxyClient import FileCatalogProxyClient
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader

__RCSID__ = "$Id$"


class FileCatalogFactory(object):
    """Factory of file catalog objects. Only exposes createCatalog() method"""

    def __init__(self):
        self.log = gLogger.getSubLogger("FileCatalogFactory")
        self.catalogPath = ""

    def createCatalog(self, catalogName, useProxy=False):
        """Create a file catalog object from its name and CS description"""
        catalogPath = getCatalogPath(catalogName)
        catalogType = gConfig.getValue(catalogPath + "/CatalogType", catalogName)
        catalogURL = gConfig.getValue(catalogPath + "/CatalogURL", "DataManagement/" + catalogType)
        optionsDict = {}
        result = gConfig.getOptionsDict(catalogPath)
        if result["OK"]:
            optionsDict = result["Value"]

        if useProxy:
            result = self.__getCatalogClass(catalogType)
            if not result["OK"]:
                return result
            catalogClass = result["Value"]
            methods = catalogClass.getInterfaceMethods()
            catalog = FileCatalogProxyClient(catalogName)
            catalog.setInterfaceMethods(methods)
            return S_OK(catalog)

        return self.__createCatalog(catalogName, catalogType, catalogURL, optionsDict)

    def __getCatalogClass(self, catalogType):
        result = ObjectLoader().loadObject("Resources.Catalog.%sClient" % catalogType)
        if not result["OK"]:
            gLogger.error("Failed to load catalog object", "%s" % result["Message"])
        return result

    def __createCatalog(self, catalogName, catalogType, catalogURL, optionsDict):

        self.log.debug("Creating %s client of type %s" % (catalogName, catalogType))

        result = self.__getCatalogClass(catalogType)
        if not result["OK"]:
            return result
        catalogClass = result["Value"]

        try:
            optionsDict["url"] = catalogURL
            catalog = catalogClass(**optionsDict)
            self.log.debug("Loaded module %sClient" % catalogType)
            return S_OK(catalog)
        except Exception as x:
            errStr = "Failed to instantiate %s()" % (catalogType)
            gLogger.exception(errStr, lException=x)
            return S_ERROR(errStr)

        # Catalog module was not loaded
        return S_ERROR("No suitable client found for %s" % catalogName)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

__RCSID__ = "$Id$"


def VmB(vmKey):
    __memScale = {"kB": 1024.0, "mB": 1024.0 * 1024.0, "KB": 1024.0, "MB": 1024.0 * 1024.0}
    __vmKeys = [
        "VmPeak:",
        "VmSize:",
        "VmLck:",
        "VmHWM:",
        "VmRSS:",
        "VmData:",
        "VmStk:",
        "VmExe:",
        "VmLib:",
        "VmPTE:",
        "VmPeak",
        "VmSize",
        "VmLck",
        "VmHWM",
        "VmRSS",
        "VmData",
        "VmStk",
        "VmExe",
        "VmLib",
        "VmPTE",
    ]
    if vmKey not in __vmKeys:
        return 0
    procFile = "/proc/%d/status" % os.getpid()
    # get pseudo file  /proc/<pid>/status
    try:
        with open(procFile) as myFile:
            value = myFile.read()
    except Exception:
        return 0.0  # non-Linux?
    # get vmKey line e.g. 'VmRSS:  9999  kB\n ...'
    i = value.index(vmKey)
    value = value[i:].split(None, 3)  # whitespace
    if len(value) < 3:
        return 0.0  # invalid format?
    # convert Vm value to bytes
    return float(value[1]) * __memScale[value[2]]



"""
   Testing the FileCatalog logic
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import unittest
import mock

import DIRAC
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

from DIRAC import S_OK, S_ERROR

__RCSID__ = "$Id $"

current_module = sys.modules[__name__]


class GenericCatalog(object):
    """Dummy catalog"""

    def __init__(self, name, nb_read, nb_read_no_lfn, nb_write, nb_write_no_lfn):

        self.w_method = []
        self.r_method = []
        self.no_lfn = []
        self.name = name

        self.__generateMethods("read", self.r_method, nb_read, nb_read_no_lfn)
        self.__generateMethods("write", self.w_method, nb_write, nb_write_no_lfn)

    def hasCatalogMethod(self, methName):
        return methName in self.w_method or methName in self.r_method

    def __generateMethods(self, mType, methodList, nb_method, nb_method_no_lfn):
        """Generates methods, read or write, and adds them to the appropriate list,
        including no_lfn if needed.
        The no_lfn methods are taken starting from the end.

        :param mType: read or write
        :param methodList: in which list to put the names
        :param nb_methods: number of methods
        :param nb_methods_no_lfn: number of these methods that should go in no lfn

        """

        no_lfn_start = nb_method - nb_method_no_lfn + 1
        for catId in range(1, nb_method + 1):
            mName = "%s%d" % (mType, catId)
            methodList.append(mName)
            if catId >= no_lfn_start:
                self.no_lfn.append(mName)

    def getInterfaceMethods(self):
        return self.r_method, self.w_method, self.no_lfn

    def generic(self, *args, **kwargs):
        """Returns a status depending on the input.
        For a normal read or write method, it looks for the catalog
        name in the LFN. If it is there, it looks at which status it is
        supposed to return: S_Error, or put the LFN in the Failed dict.
        """

        successful = {}
        failed = {}

        if self.call in self.no_lfn:
            if not args:
                return S_OK("yeah")
            ret = args[0]
            if self.name in ret:
                return S_ERROR("%s.%s did not go well" % (self.name, self.call))
            else:
                return S_OK("yeah")

        lfns = args[0]

        for lfn in lfns:
            lfnSplit = lfn.split("/")
            try:
                idName = lfnSplit.index(self.name)
                retType = lfnSplit[idName + 1]
                if retType == "Error":
                    return S_ERROR("%s.%s did not go well" % (self.name, self.call))
                elif retType == "Failed":
                    failed[lfn] = "%s.%s failed for %s" % (self.name, self.call, lfn)
            except ValueError:
                successful[lfn] = "yeah"

        return S_OK({"Successful": successful, "Failed": failed})

    def __getattr__(self, meth):
        self.call = meth
        return self.generic


def mock_fc_getSelectedCatalogs(self, desiredCatalogs):
    """Mock the getSelectedCatalogs method
    The name of the catalog should contain the following info, separated by '_':
      * the name of the catalog
      * True or False if it is a Master
      * True or False for Read
      * True or False for Write
      * nb of read op
      * nb of read no lfn
      * nb of write op
      * nb of write no lfn
    """

    for catalogDescription in desiredCatalogs:
        name, master, read, write, nb_read, nb_read_no_lfn, nb_write, nb_write_no_lfn = catalogDescription.split("_")
        master = eval(master)
        read = eval(read)
        write = eval(write)
        nb_read = eval(nb_read)
        nb_read_no_lfn = eval(nb_read_no_lfn)
        nb_write = eval(nb_write)
        nb_write_no_lfn = eval(nb_write_no_lfn)

        obj = GenericCatalog(name, nb_read, nb_read_no_lfn, nb_write, nb_write_no_lfn)

        if read:
            self.readCatalogs.append((name, obj, master))
        if write:
            self.writeCatalogs.append((name, obj, master))

    return S_OK()


def mock_fc_getEligibleCatalogs(self):
    """We return an object that always returns True
    if we ask whether an item is in it
    """

    class mockList(object):
        def __contains__(self, item):
            return True

    x = mockList()
    return S_OK(x)


def writeList(count, reverse=None):
    """If reverse is none, returns write1, ...., write<count>
    if reverse is set, returns a list with <count> elements backward from read<reverse>
    """
    if reverse:
        return ["write%s" % i for i in range(reverse, reverse - count, -1)]
    return ["write%s" % i for i in range(1, count + 1)]


def readList(count, reverse=None):
    """If reverse is none, returns read1, ...., read<count>
    if reverse is set, returns a list with <count> elements backward from read<reverse>
    """
    if reverse:
        return ["read%s" % i for i in range(reverse, reverse - count, -1)]

    return ["read%s" % i for i in range(1, count + 1)]


class TestInitialization(unittest.TestCase):
    """Tests the logic of the init mechanism"""

    @mock.patch.object(
        DIRAC.Resources.Catalog.FileCatalog.FileCatalog,
        "_getSelectedCatalogs",
        side_effect=mock_fc_getSelectedCatalogs,
        autospec=True,
    )  # autospec is for the binding of the method...
    @mock.patch.object(
        DIRAC.Resources.Catalog.FileCatalog.FileCatalog,
        "_getEligibleCatalogs",
        side_effect=mock_fc_getEligibleCatalogs,
        autospec=True,
    )  # autospec is for the binding of the method...
    def test_01_init(self, mk_getSelectedCatalogs, mk_getEligibleCatalogs):
        """Check logic of init"""

        # We should not be able to have 2 masters
        twoMastersFc = FileCatalog(catalogs=["c1_True_True_True_5_2_2_0", "c2_True_True_True_5_2_2_0"])
        self.assertTrue(not twoMastersFc.isOK())

        # One master should be ok
        oneMasterFc = FileCatalog(catalogs=["c1_True_True_True_2_0_2_2", "c2_False_True_True_3_1_4_2"])
        self.assertTrue(oneMasterFc.isOK())

        # With a master, the write method should be the method of the master
        self.assertEqual(sorted(oneMasterFc.write_methods), writeList(2))
        # The read methods and no_lfn should be from all catalogs
        self.assertEqual(sorted(oneMasterFc.ro_methods), readList(3))

        # The no_lfns methods are from everywhere
        # write1 and write2 from c1
        # write3, write4, read3 from c2
        self.assertEqual(sorted(oneMasterFc.no_lfn_methods), sorted(readList(1, reverse=3) + writeList(4)))

        # No master should be ok
        noMasterFc = FileCatalog(catalogs=["c1_False_True_True_2_0_2_0", "c2_False_True_True_3_0_4_0"])
        self.assertTrue(oneMasterFc.isOK())
        # With no master, the write method should be from all catalogs
        self.assertEqual(sorted(noMasterFc.write_methods), writeList(4))
        # The read methods and no_lfn should be from all catalogs
        self.assertEqual(sorted(noMasterFc.ro_methods), readList(3))


class TestWrite(unittest.TestCase):
    """Tests of the w_execute method"""

    @mock.patch.object(
        DIRAC.Resources.Catalog.FileCatalog.FileCatalog,
        "_getSelectedCatalogs",
        side_effect=mock_fc_getSelectedCatalogs,
        autospec=True,
    )  # autospec is for the binding of the method...
    @mock.patch.object(
        DIRAC.Resources.Catalog.FileCatalog.FileCatalog,
        "_getEligibleCatalogs",
        side_effect=mock_fc_getEligibleCatalogs,
        autospec=True,
    )  # autospec is for the binding of the method...
    def test_01_Normal(self, mk_getSelectedCatalogs, mk_getEligibleCatalogs):
        """Test behavior with one master and only standard write methods"""

        fc = FileCatalog(catalogs=["c1_True_True_True_2_0_2_0", "c2_False_True_True_3_0_1_0"])

        # Test a write method which is not in the master catalog
        with self.assertRaises(AttributeError):
            fc.write4("/lhcb/toto")

        # Test a write method which works for everybody
        lfn = "/lhcb/toto"
        res = fc.write1(lfn)
        self.assertTrue(res["OK"])
        self.assertTrue(lfn in res["Value"]["Successful"])
        self.assertEqual(sorted(["c1", "c2"]), sorted(res["Value"]["Successful"][lfn]))
        self.assertTrue(not res["Value"]["Failed"])

        # Test a write method that only the master has
        lfn = "/lhcb/toto"
        res = fc.write2(lfn)
        self.assertTrue(res["OK"])
        self.assertTrue(lfn in res["Value"]["Successful"])
        self.assertEqual(["c1"], sorted(res["Value"]["Successful"][lfn]))
        self.assertTrue(not res["Value"]["Failed"])

        # Test a write method that makes an error for master
        # We should get an error
        lfn = "/lhcb/c1/Error"
        res = fc.write1(lfn)
        self.assertTrue(not res["OK"])

        # Test a write method that fails for master
        # The lfn should be in failed and only attempted for the master
        lfn = "/lhcb/c1/Failed"
        res = fc.write1(lfn)
        self.assertTrue(res["OK"])
        self.assertTrue(not res["Value"]["Successful"])
        self.assertEqual(["c1"], sorted(res["Value"]["Failed"][lfn]))

        # Test a write method that makes an error for non master
        # The lfn should be in failed for non master and successful for the master
        lfn = "/lhcb/c2/Error"
        res = fc.write1(lfn)
        self.assertTrue(res["OK"])
        self.assertEqual(["c1"], sorted(res["Value"]["Successful"][lfn]))
        self.assertEqual(["c2"], sorted(res["Value"]["Failed"][lfn]))

        # Test a write method that fails for non master
        # The lfn should be in failed for non master and successful for the master
        lfn = "/lhcb/c2/Failed"
        res = fc.write1(lfn)
        self.assertTrue(res["OK"])
        self.assertEqual(["c1"], sorted(res["Value"]["Successful"][lfn]))
        self.assertEqual(["c2"], sorted(res["Value"]["Failed"][lfn]))

    @mock.patch.object(
        DIRAC.Resources.Catalog.FileCatalog.FileCatalog,
        "_getSelectedCatalogs",
        side_effect=mock_fc_getSelectedCatalogs,
        autospec=True,
    )  # autospec is for the binding of the method...
    @mock.patch.object(
        DIRAC.Resources.Catalog.FileCatalog.FileCatalog,
        "_getEligibleCatalogs",
        side_effect=mock_fc_getEligibleCatalogs,
        autospec=True,
    )  # autospec is for the binding of the method...
    def test_02_condParser(self, mk_getSelectedCatalogs, mk_getEligibleCatalogs):
        """Test behavior of write methode when using FCConditionParser"""

        fc = FileCatalog(
            catalogs=["c1_True_True_True_2_0_2_0", "c2_False_True_True_3_0_1_0", "c3_False_True_True_3_0_1_0"]
        )

        # No condition for c3, so it should always pass
        fcConditions = {"c1": "Filename=find('c1_pass')", "c2": "Filename=find('c2_pass')"}

        # Everything pass everywhere
        lfn1 = "/lhcb/c1_pass/c2_pass/lfn1"
        lfn2 = "/lhcb/c1_pass/c2_pass/lfn2"
        res = fc.write1([lfn1, lfn2], fcConditions=fcConditions)
        self.assertTrue(res["OK"])
        self.assertEqual(sorted(res["Value"]["Successful"]), sorted([lfn1, lfn2]))
        self.assertEqual(sorted(res["Value"]["Successful"][lfn1]), sorted(["c1", "c2", "c3"]))
        self.assertEqual(sorted(res["Value"]["Successful"][lfn2]), sorted(["c1", "c2", "c3"]))
        self.assertTrue(not res["Value"]["Failed"])

        # Everything pass for the master, only lfn2 for c2
        lfn1 = "/lhcb/c1_pass/lfn1"
        lfn2 = "/lhcb/c1_pass/c2_pass/lfn2"
        res = fc.write1([lfn1, lfn2], fcConditions=fcConditions)
        self.assertTrue(res["OK"])
        self.assertEqual(sorted(res["Value"]["Successful"]), sorted([lfn1, lfn2]))
        self.assertEqual(sorted(res["Value"]["Successful"][lfn1]), ["c1", "c3"])
        self.assertEqual(sorted(res["Value"]["Successful"][lfn2]), sorted(["c1", "c2", "c3"]))
        self.assertTrue(not res["Value"]["Failed"])

        # One is not valid for the master, so we do nothing
        lfn1 = "/lhcb/c2_pass/lfn1"
        lfn2 = "/lhcb/c1_pass/c2_pass/lfn2"
        res = fc.write1([lfn1, lfn2], fcConditions=fcConditions)
        self.assertTrue(not res["OK"])

    @mock.patch.object(
        DIRAC.Resources.Catalog.FileCatalog.FileCatalog,
        "_getSelectedCatalogs",
        side_effect=mock_fc_getSelectedCatalogs,
        autospec=True,
    )  # autospec is for the binding of the method...
    @mock.patch.object(
        DIRAC.Resources.Catalog.FileCatalog.FileCatalog,
        "_getEligibleCatalogs",
        side_effect=mock_fc_getEligibleCatalogs,
        autospec=True,
    )  # autospec is for the binding of the method...
    def test_03_noLFN(self, mk_getSelectedCatalogs, mk_getEligibleCatalogs):
        """Test the no_lfn methods"""

        fc = FileCatalog(catalogs=["c1_True_True_True_2_0_2_1", "c2_False_True_True_3_0_2_1"])

        # all good
        res = fc.write2("/lhcb/toto")
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"], "yeah")

        # Fail in the master
        res = fc.write2("/lhcb/c1")
        self.assertTrue(not res["OK"])
        self.assertTrue("Value" not in res)

        # Fail in the non master
        res = fc.write2("/lhcb/c2")
        self.assertTrue(res["OK"])
        self.assertTrue("Value" in res)
        self.assertEqual(res["Value"], "yeah")


class TestRead(unittest.TestCase):
    """Tests of the w_execute method"""

    @mock.patch.object(
        DIRAC.Resources.Catalog.FileCatalog.FileCatalog,
        "_getSelectedCatalogs",
        side_effect=mock_fc_getSelectedCatalogs,
        autospec=True,
    )  # autospec is for the binding of the method...
    @mock.patch.object(
        DIRAC.Resources.Catalog.FileCatalog.FileCatalog,
        "_getEligibleCatalogs",
        side_effect=mock_fc_getEligibleCatalogs,
        autospec=True,
    )  # autospec is for the binding of the method...
    def test_01_oneMasterNormal(self, mk_getSelectedCatalogs, mk_getEligibleCatalogs):
        """Test behavior with one master and only standard read methods"""

        fc = FileCatalog(catalogs=["c1_True_True_True_2_0_2_0", "c2_False_True_True_3_0_1_0"])

        # Test a write method which is not in the master catalog
        with self.assertRaises(AttributeError):
            fc.write4("/lhcb/toto")

        # Test a write method which works for everybody
        lfn = "/lhcb/toto"
        res = fc.write1(lfn)
        self.assertTrue(res["OK"])
        self.assertTrue(lfn in res["Value"]["Successful"])
        self.assertEqual(sorted(["c1", "c2"]), sorted(res["Value"]["Successful"][lfn]))
        self.assertTrue(not res["Value"]["Failed"])

        # Test a write method that only the master has
        lfn = "/lhcb/toto"
        res = fc.write2(lfn)
        self.assertTrue(res["OK"])
        self.assertTrue(lfn in res["Value"]["Successful"])
        self.assertEqual(["c1"], sorted(res["Value"]["Successful"][lfn]))
        self.assertTrue(not res["Value"]["Failed"])

        # Test a write method that makes an error for master
        # We should get an error
        lfn = "/lhcb/c1/Error"
        res = fc.write1(lfn)
        self.assertTrue(not res["OK"])

        # Test a write method that fails for master
        # The lfn should be in failed and only attempted for the master
        lfn = "/lhcb/c1/Failed"
        res = fc.write1(lfn)
        self.assertTrue(res["OK"])
        self.assertTrue(not res["Value"]["Successful"])
        self.assertEqual(["c1"], sorted(res["Value"]["Failed"][lfn]))

        # Test a write method that makes an error for non master
        # The lfn should be in failed for non master and successful for the master
        lfn = "/lhcb/c2/Error"
        res = fc.write1(lfn)
        self.assertTrue(res["OK"])
        self.assertEqual(["c1"], sorted(res["Value"]["Successful"][lfn]))
        self.assertEqual(["c2"], sorted(res["Value"]["Failed"][lfn]))

        # Test a write method that fails for non master
        # The lfn should be in failed for non master and successful for the master
        lfn = "/lhcb/c2/Failed"
        res = fc.write1(lfn)
        self.assertTrue(res["OK"])
        self.assertEqual(["c1"], sorted(res["Value"]["Successful"][lfn]))
        self.assertEqual(["c2"], sorted(res["Value"]["Failed"][lfn]))


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestInitialization)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestWrite))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestRead))

    unittest.TextTestRunner(verbosity=2).run(suite)



""" Test class for AccountingDB
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access

# imports
import unittest
from mock import MagicMock

import DIRAC.AccountingSystem.DB.AccountingDB as moduleTested


class TestCase(unittest.TestCase):
    """Base class for the AccountingDB test cases"""

    def setUp(self):

        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.AccountingDB

        self.moduleTested.DB = MagicMock()
        self.moduleTested.DB.fullname = "AccountingDB"

        mock_getDatabaseSection = MagicMock()
        mock_getDatabaseSection.return_value = "/Systems/AccountingSystem/Certification/Databases/AccountingDB"
        self.moduleTested.DB.getDatabaseSection = mock_getDatabaseSection

        self.moduleTested.AccountingDB.fullname = MagicMock()

        mock_getCSOption = MagicMock()
        mock_getCSOption.return_value = "10"
        self.moduleTested.AccountingDB.getCSOption = mock_getCSOption

        mock_ThreadPool = MagicMock()
        self.moduleTested.ThreadPool = mock_ThreadPool

        mock_createTables = MagicMock()
        mock_createTables.return_value = {"OK": True}
        self.moduleTested.AccountingDB._createTables = mock_createTables

        mock_loadCatalogFromDB = MagicMock()
        mock_loadCatalogFromDB.return_value = {"OK": True}
        self.moduleTested.AccountingDB._AccountingDB__loadCatalogFromDB = mock_loadCatalogFromDB

        mock_registerTypes = MagicMock()
        mock_registerTypes.return_value = {"OK": True}
        self.moduleTested.AccountingDB._AccountingDB__registerTypes = mock_registerTypes
        mock_log = MagicMock()
        mock_log.return_value = {"OK": True}
        self.moduleTested.AccountingDB.log = mock_log


class MakeQuery(TestCase):
    """testing making a query"""

    def query(self, cmd, conn):  # pylint: disable=no-self-use,unused-argument
        """Because we are not able to execute the query, the method returns the query"""
        return cmd

    ################################################################################
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class"""

        module = self.testClass()
        self.assertEqual("AccountingDB", module.__class__.__name__)

    def test_calculateBuckets(self):
        """
        Test the method used to calculate the bucket size
        """

        module = self.testClass()
        module.dbCatalog = {
            "LHCb-Certification_DataOperation": {
                "definition": {
                    "keys": [
                        ("OperationType", "VARCHAR(32)"),
                        ("User", "VARCHAR(32)"),
                        ("ExecutionSite", "VARCHAR(32)"),
                        ("Source", "VARCHAR(32)"),
                        ("Destination", "VARCHAR(32)"),
                        ("Protocol", "VARCHAR(32)"),
                        ("FinalStatus", "VARCHAR(32)"),
                    ],
                    "values": [
                        ("TransferSize", "BIGINT UNSIGNED"),
                        ("TransferTime", "FLOAT"),
                        ("RegistrationTime", "FLOAT"),
                        ("TransferOK", "INT UNSIGNED"),
                        ("TransferTotal", "INT UNSIGNED"),
                        ("RegistrationOK", "INT UNSIGNED"),
                        ("RegistrationTotal", "INT UNSIGNED"),
                    ],
                },
                "bucketFields": [
                    "OperationType",
                    "User",
                    "ExecutionSite",
                    "Source",
                    "Destination",
                    "Protocol",
                    "FinalStatus",
                    "TransferSize",
                    "TransferTime",
                    "RegistrationTime",
                    "TransferOK",
                    "TransferTotal",
                    "RegistrationOK",
                    "RegistrationTotal",
                    "entriesInBucket",
                    "startTime",
                    "bucketLength",
                ],
                "keys": ["OperationType", "User", "ExecutionSite", "Source", "Destination", "Protocol", "FinalStatus"],
                "typeFields": [
                    "OperationType",
                    "User",
                    "ExecutionSite",
                    "Source",
                    "Destination",
                    "Protocol",
                    "FinalStatus",
                    "TransferSize",
                    "TransferTime",
                    "RegistrationTime",
                    "TransferOK",
                    "TransferTotal",
                    "RegistrationOK",
                    "RegistrationTotal",
                    "startTime",
                    "endTime",
                ],
                "values": [
                    "TransferSize",
                    "TransferTime",
                    "RegistrationTime",
                    "TransferOK",
                    "TransferTotal",
                    "RegistrationOK",
                    "RegistrationTotal",
                ],
                "dataTimespan": 0,
            }
        }

        module.dbBucketsLength["LHCb-Certification_DataOperation"] = [
            (259200, 900),
            (691200, 3600),
            (15552000, 86400),
            (31104000, 604800),
        ]
        retVal = module.calculateBuckets("LHCb-Certification_DataOperation", 1495328400, 1495328400)[0][0]
        self.assertTrue(retVal)

        retVal = module.calculateBuckets("LHCb-Certification_DataOperation", 1497964315, 1497964315)[0][0]
        self.assertTrue(retVal)
        self.assertTrue(retVal)

        retVal = module.calculateBuckets("LHCb-Certification_DataOperation", 1495414800, 1495414800)[0][0]
        self.assertTrue(retVal)

        retVal = module.calculateBuckets("LHCb-Certification_DataOperation", 1498047038, 1498047038)[0][0]
        self.assertTrue(retVal)

    def test__queryType1(self):
        """
        Test the query creation for a given condition
        """
        module = self.testClass()
        module.dbCatalog = {
            "LHCb-Certification_DataOperation": {
                "definition": {
                    "keys": [
                        ("OperationType", "VARCHAR(32)"),
                        ("User", "VARCHAR(32)"),
                        ("ExecutionSite", "VARCHAR(32)"),
                        ("Source", "VARCHAR(32)"),
                        ("Destination", "VARCHAR(32)"),
                        ("Protocol", "VARCHAR(32)"),
                        ("FinalStatus", "VARCHAR(32)"),
                    ],
                    "values": [
                        ("TransferSize", "BIGINT UNSIGNED"),
                        ("TransferTime", "FLOAT"),
                        ("RegistrationTime", "FLOAT"),
                        ("TransferOK", "INT UNSIGNED"),
                        ("TransferTotal", "INT UNSIGNED"),
                        ("RegistrationOK", "INT UNSIGNED"),
                        ("RegistrationTotal", "INT UNSIGNED"),
                    ],
                },
                "bucketFields": [
                    "OperationType",
                    "User",
                    "ExecutionSite",
                    "Source",
                    "Destination",
                    "Protocol",
                    "FinalStatus",
                    "TransferSize",
                    "TransferTime",
                    "RegistrationTime",
                    "TransferOK",
                    "TransferTotal",
                    "RegistrationOK",
                    "RegistrationTotal",
                    "entriesInBucket",
                    "startTime",
                    "bucketLength",
                ],
                "keys": ["OperationType", "User", "ExecutionSite", "Source", "Destination", "Protocol", "FinalStatus"],
                "typeFields": [
                    "OperationType",
                    "User",
                    "ExecutionSite",
                    "Source",
                    "Destination",
                    "Protocol",
                    "FinalStatus",
                    "TransferSize",
                    "TransferTime",
                    "RegistrationTime",
                    "TransferOK",
                    "TransferTotal",
                    "RegistrationOK",
                    "RegistrationTotal",
                    "startTime",
                    "endTime",
                ],
                "values": [
                    "TransferSize",
                    "TransferTime",
                    "RegistrationTime",
                    "TransferOK",
                    "TransferTotal",
                    "RegistrationOK",
                    "RegistrationTotal",
                ],
                "dataTimespan": 0,
            }
        }

        module.dbBucketsLength["LHCb-Certification_DataOperation"] = [
            (259200, 900),
            (691200, 3600),
            (15552000, 86400),
            (31104000, 604800),
        ]
        module._query = self.query
        startTime = 1495324800
        endTime = 1497960715
        retVal = module.calculateBuckets("LHCb-Certification_DataOperation", startTime + 3600, startTime + 3600)[0][0]
        self.assertTrue(retVal)
        expectedStartTime = retVal

        retVal = module.calculateBuckets("LHCb-Certification_DataOperation", endTime + 3600, endTime + 3600)[0][0]
        self.assertTrue(retVal)
        expectedEndTime = retVal

        expectedQuery = (
            "SELECT `ac_key_LHCb-Certification_DataOperation_Source`.`value`, \
`ac_bucket_LHCb-Certification_DataOperation`.`startTime`, \
`ac_bucket_LHCb-Certification_DataOperation`.`bucketLength`, \
SUM(`ac_bucket_LHCb-Certification_DataOperation`.`TransferOK`), \
SUM(`ac_bucket_LHCb-Certification_DataOperation`.`TransferTotal`)-\
SUM(`ac_bucket_LHCb-Certification_DataOperation`.`TransferOK`) \
FROM `ac_bucket_LHCb-Certification_DataOperation`, \
`ac_key_LHCb-Certification_DataOperation_Source` \
WHERE `ac_bucket_LHCb-Certification_DataOperation`.`startTime` >= %s \
AND `ac_bucket_LHCb-Certification_DataOperation`.`startTime` <= %s \
AND `ac_bucket_LHCb-Certification_DataOperation`.`Source` = `ac_key_LHCb-Certification_DataOperation_Source`.`id` \
GROUP BY startTime, `ac_key_LHCb-Certification_DataOperation_Source`.Value, bucketlength \
ORDER BY startTime"
            % (expectedStartTime, expectedEndTime)
        )
        retVal = module._AccountingDB__queryType(
            "LHCb-Certification_DataOperation",  # pylint: disable=no-member
            startTime,
            endTime,
            (
                "%s, %s, %s, SUM(%s), SUM(%s)-SUM(%s)",
                ["Source", "startTime", "bucketLength", "TransferOK", "TransferTotal", "TransferOK"],
            ),
            {},
            ("%s, %s", ["startTime", "Source"]),
            ("%s", ["startTime"]),
            "bucket",
        )

        self.assertTrue(retVal)
        self.assertEqual(retVal, expectedQuery)

    def test_queryType2(self):
        """Test the query creation for a given condition"""
        module = self.testClass()
        module.dbCatalog = {
            "LHCb-Certification_DataOperation": {
                "definition": {
                    "keys": [
                        ("OperationType", "VARCHAR(32)"),
                        ("User", "VARCHAR(32)"),
                        ("ExecutionSite", "VARCHAR(32)"),
                        ("Source", "VARCHAR(32)"),
                        ("Destination", "VARCHAR(32)"),
                        ("Protocol", "VARCHAR(32)"),
                        ("FinalStatus", "VARCHAR(32)"),
                    ],
                    "values": [
                        ("TransferSize", "BIGINT UNSIGNED"),
                        ("TransferTime", "FLOAT"),
                        ("RegistrationTime", "FLOAT"),
                        ("TransferOK", "INT UNSIGNED"),
                        ("TransferTotal", "INT UNSIGNED"),
                        ("RegistrationOK", "INT UNSIGNED"),
                        ("RegistrationTotal", "INT UNSIGNED"),
                    ],
                },
                "bucketFields": [
                    "OperationType",
                    "User",
                    "ExecutionSite",
                    "Source",
                    "Destination",
                    "Protocol",
                    "FinalStatus",
                    "TransferSize",
                    "TransferTime",
                    "RegistrationTime",
                    "TransferOK",
                    "TransferTotal",
                    "RegistrationOK",
                    "RegistrationTotal",
                    "entriesInBucket",
                    "startTime",
                    "bucketLength",
                ],
                "keys": ["OperationType", "User", "ExecutionSite", "Source", "Destination", "Protocol", "FinalStatus"],
                "typeFields": [
                    "OperationType",
                    "User",
                    "ExecutionSite",
                    "Source",
                    "Destination",
                    "Protocol",
                    "FinalStatus",
                    "TransferSize",
                    "TransferTime",
                    "RegistrationTime",
                    "TransferOK",
                    "TransferTotal",
                    "RegistrationOK",
                    "RegistrationTotal",
                    "startTime",
                    "endTime",
                ],
                "values": [
                    "TransferSize",
                    "TransferTime",
                    "RegistrationTime",
                    "TransferOK",
                    "TransferTotal",
                    "RegistrationOK",
                    "RegistrationTotal",
                ],
                "dataTimespan": 0,
            }
        }

        module.dbBucketsLength["LHCb-Certification_DataOperation"] = [
            (259200, 900),
            (691200, 3600),
            (15552000, 86400),
            (31104000, 604800),
        ]
        module._query = self.query
        startTime = 1495411200
        endTime = 1498043438
        retVal = module.calculateBuckets("LHCb-Certification_DataOperation", startTime + 3600, startTime + 3600)[0][0]
        self.assertTrue(retVal)
        expectedStartTime = retVal

        retVal = module.calculateBuckets("LHCb-Certification_DataOperation", endTime + 3600, endTime + 3600)[0][0]
        self.assertTrue(retVal)
        expectedEndTime = retVal

        expectedQuery = (
            "SELECT `ac_key_LHCb-Certification_DataOperation_Source`.`value`, \
`ac_bucket_LHCb-Certification_DataOperation`.`startTime`, \
`ac_bucket_LHCb-Certification_DataOperation`.`bucketLength`, \
SUM(`ac_bucket_LHCb-Certification_DataOperation`.`TransferOK`), \
SUM(`ac_bucket_LHCb-Certification_DataOperation`.`TransferTotal`)-\
SUM(`ac_bucket_LHCb-Certification_DataOperation`.`TransferOK`) \
FROM `ac_bucket_LHCb-Certification_DataOperation`, \
`ac_key_LHCb-Certification_DataOperation_Source` \
WHERE `ac_bucket_LHCb-Certification_DataOperation`.`startTime` >= %s \
AND `ac_bucket_LHCb-Certification_DataOperation`.`startTime` <= %s \
AND `ac_bucket_LHCb-Certification_DataOperation`.`Source` = `ac_key_LHCb-Certification_DataOperation_Source`.`id` \
GROUP BY startTime, `ac_key_LHCb-Certification_DataOperation_Source`.Value, bucketlength \
ORDER BY startTime"
            % (expectedStartTime, expectedEndTime)
        )

        retVal = module._AccountingDB__queryType(
            "LHCb-Certification_DataOperation",  # pylint: disable=no-member
            startTime,
            endTime,
            (
                "%s, %s, %s, SUM(%s), SUM(%s)-SUM(%s)",
                ["Source", "startTime", "bucketLength", "TransferOK", "TransferTotal", "TransferOK"],
            ),
            {},
            ("%s, %s", ["startTime", "Source"]),
            ("%s", ["startTime"]),
            "bucket",
        )

        self.assertTrue(retVal)
        self.assertEqual(retVal, expectedQuery)


#############################################################################
# Test Suite run
#############################################################################


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(MakeQuery))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)
